from typing import Optional, List, Union

import numpy as np
import vtk
import cadquery as cq

import vmi

vmi.CS4x4 = None


class CS4x4:
    """
    坐标系，以4x4矩阵表示的局部坐标系或坐标变换
    """

    def __init__(self, ary=None, axis0: Union[np.ndarray, List[float]] = None,
                 axis1: Union[np.ndarray, List[float]] = None, axis2: Union[np.ndarray, List[float]] = None,
                 origin: Union[np.ndarray, List[float]] = None):
        """
        局部坐标系

        :param axis0: 3x1第一坐标轴，转为齐次坐标构成矩阵的第一行
        :param axis1: 3x1第二坐标轴，转为齐次坐标构成矩阵的第二行
        :param axis2: 3x1第三坐标轴，转为齐次坐标构成矩阵的第三行
        :param origin: 3x1坐标原点，转为齐次坐标构成矩阵的第四行
        """
        self._ary = np.identity(4)
        if ary is not None:
            self.setArray4x4(ary)
        if axis0 is not None:
            self.setAxis(0, axis0)
        if axis1 is not None:
            self.setAxis(1, axis1)
        if axis2 is not None:
            self.setAxis(2, axis2)
        if origin is not None:
            self.setOrigin(origin)

    def __repr__(self):
        return self.array4x4().__repr__()

    def copy(self) -> vmi.CS4x4:
        """
        复制实例

        :return: vmi.CS4x4 - 具有相同矩阵的实例
        """
        cs = CS4x4()
        cs._ary = self._ary.copy()
        return cs

    def workplane(self, local_origin: Union[np.ndarray, List[float]] = (0, 0, 0)) -> cq.Workplane:
        return cq.Workplane(cq.Plane(self.mpt(local_origin).tolist(), self.axis(0).tolist(), self.axis(2).tolist()))

    def array4x4(self) -> np.ndarray:
        """返回矩阵"""
        return self._ary

    def setArray4x4(self, ary: Union[np.ndarray, List[float]] = None):
        """设置矩阵"""
        self._ary = ary.copy()

    def ary4x4inv(self) -> np.ndarray:
        """返回当前矩阵的逆矩阵"""
        return np.linalg.inv(self.array4x4())

    def matrix4x4(self, mat: vtk.vtkMatrix4x4 = None) -> vtk.vtkMatrix4x4:
        """返回当前矩阵的vtk矩阵"""
        ary = self.array4x4()
        matrix = vtk.vtkMatrix4x4()
        for j in range(4):
            for i in range(4):
                matrix.SetElement(j, i, ary[i, j])
        return matrix

    def setMatrix4x4(self, mat: vtk.vtkMatrix4x4 = None):
        """设置vtkMatrix4x4矩阵"""
        ary = np.zeros((4, 4))
        for j in range(4):
            for i in range(4):
                ary[i, j] = mat.GetElement(j, i)

    def mat4x4inv(self) -> vtk.vtkMatrix4x4:
        """返回当前矩阵的vtkMatrix4x4逆矩阵"""
        ary = self.ary4x4inv()
        matrix = vtk.vtkMatrix4x4()
        for j in range(4):
            for i in range(4):
                matrix.SetElement(j, i, ary[i, j])
        return matrix

    def inv(self) -> vmi.CS4x4:
        """返回逆变换"""
        cs = CS4x4()
        cs.setArray4x4(self.ary4x4inv())
        return cs

    def axis(self, i: int) -> np.ndarray:
        """返回第i坐标轴"""
        return self._ary[i][:3]

    def setAxis(self, i: int, arg: Union[np.ndarray, List[float]] = None):
        """设置第i坐标轴"""
        for j in range(3):
            self._ary[i][j] = arg[j]

    def origin(self) -> np.ndarray:
        """返回坐标原点"""
        return self._ary[3][:3]

    def setOrigin(self, arg: Union[np.ndarray, List[float]] = None):
        """设置坐标原点"""
        for j in range(3):
            self._ary[3][j] = arg[j]

    def orthogonalize(self, i_list: List[int]) -> None:
        """
        按照次序列表正交化坐标轴，例如i_list=[0, 2]表示先用z轴y轴正交化x轴，再用x轴y轴正交化z轴

        :param i_list: 坐标轴次序列表
        :return: None
        """
        for i in i_list:
            self.setAxis(i, np.cross(self.axis((i + 1) % 3), self.axis((i + 2) % 3)))

    def normalize(self, i_list: List[int] = None) -> None:
        """
        按照次序列表单位化坐标轴，例如i_list=[0, 2]表示将x轴z轴单位化

        :param i_list: 坐标轴次序列表
        :return: None
        """
        if i_list is None:
            i_list = [0, 1, 2]
        for i in i_list:
            self.setAxis(i, self.axis(i) / np.linalg.norm(self.axis(i)))

    def mpt(self, pt: Union[np.ndarray, List[float]], origin: Union[np.ndarray, List[float]] = None,
            other: Optional[vmi.CS4x4] = None) -> np.ndarray:
        """
        坐标系变换：点pt由当前坐标系变换到世界坐标系，如果设置了other则变换到other坐标系
        坐标变换：世界坐标系中的点pt经过当前变换得到新点，如果设置了origin则在此次计算中原点暂时设为origin

        :param pt: 3x1点坐标
        :param origin: 3x1坐标原点
        :param other: 坐标系
        :return: 变换后坐标
        """
        origin_ = self.origin()
        if origin is not None:
            self.setOrigin(origin)

        pt = np.append(np.array(pt[:3]), 1)
        pt = pt @ self.array4x4()
        if other is not None:
            pt = pt @ other.ary4x4inv()

        if origin is not None:
            self.setOrigin(origin_)
        return pt[:3]

    def mvt(self, vt: Union[np.ndarray, List[float]], other: Optional[vmi.CS4x4] = None) -> np.ndarray:
        """
        坐标系变换：向量vt由当前坐标系变换到世界坐标系，如果设置了other则变换到other坐标系
        坐标变换：世界坐标系中的向量vt经过当前变换，获得新向量

        :param vt: 3x1向量
        :param other: 坐标系
        :return: 变换后坐标
        """
        vt = np.append(np.array(vt[:3]), 0)
        vt = vt @ self.array4x4()
        if other is not None:
            vt = vt @ other.ary4x4inv()
        return vt[:3]

    def mcs(self, cs: vmi.CS4x4):
        """
        坐标系变换：坐标系cs由当前坐标系变换到世界坐标系，如果设置了other则变换到other坐标系
        坐标变换：世界坐标系中的局部坐标系cs经过当前变换得到新坐标系
        """
        cs_ = vmi.CS4x4()
        cs_.setArray4x4(cs.array4x4() @ self.array4x4())
        return cs_

    def translate(self, vt: Union[np.ndarray, List[float]]) -> vmi.CS4x4:
        """
        串接平移变换
        :param vt: 3x1平移变换
        :return: 新坐标系
        """
        mat = np.array([[1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [vt[0], vt[1], vt[2], 1]])
        self.setArray4x4(self.array4x4() @ mat)
        return self

    def rotate(self, angle: float, axis: Union[np.ndarray, List[float]],
               origin: Union[np.ndarray, List[float]] = None) -> vmi.CS4x4:
        """
        串接旋转变换
        https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions

        :param angle: 旋转角度（deg）
        :param axis: 3x1旋转轴向量
        :param origin: 3x1旋转轴经过的一点
        :return: 新坐标系
        """
        axis = np.array(axis[:3])
        if np.linalg.norm(axis) == 0:
            raise np.linalg.LinAlgError('旋转轴向错误')

        if origin is None:
            origin = self.origin()

        t = vtk.vtkTransform()
        t.Translate(origin[0], origin[1], origin[2])
        t.RotateWXYZ(angle, axis[0], axis[1], axis[2])
        t.Translate(-origin[0], -origin[1], -origin[2])

        m = vtk.vtkMatrix4x4()
        t.GetMatrix(m)

        ary = np.zeros((4, 4))
        for j in range(4):
            for i in range(4):
                ary[i, j] = m.GetElement(j, i)
        return CS4x4(self.array4x4() @ ary)

    def reflect(self, normal: np.ndarray, origin: Union[np.ndarray, List[float]] = None) -> vmi.CS4x4:
        """
        串接反射变换
        https://en.wikipedia.org/wiki/Reflection_(mathematics)#Reflection_through_a_hyperplane_in_n_dimensions

        :param normal: 反射面法向量
        :param origin: 反射面经过的一点
        :return: 新坐标系
        """
        normal = np.array(normal[:3])
        if np.linalg.norm(normal) == 0:
            raise np.linalg.LinAlgError('反射法向错误')

        mat = np.identity(3) - 2 / np.linalg.norm(normal) ** 2 * np.outer(normal, normal)
        mat = np.hstack((np.vstack((mat, [0, 0, 0])), ([0], [0], [0], [1])))

        cs = self.copy()
        if origin is None:
            cs.setArray4x4(cs.array4x4() @ mat)
        else:
            origin = np.array(origin[:3])
            cs.translate(-origin)
            cs.setArray4x4(cs.array4x4() @ mat)
            cs.translate(origin)
        return cs

    def scale(self, factor: float, origin: Union[np.ndarray, List[float]] = None) -> vmi.CS4x4:
        """
        串接缩放变换

        :param factor: 缩放比例
        :param origin: 缩放投影中心
        :return: 新坐标系
        """
        mat = np.identity(3) * np.array(factor[:3])
        mat = np.hstack((np.vstack((mat, [0, 0, 0])), ([0], [0], [0], [1])))

        cs = self.copy()
        if origin is None:
            cs.setArray4x4(cs.array4x4() @ mat)
        else:
            origin = np.array(origin[:3])
            cs.translate(-origin)
            cs.setArray4x4(cs.array4x4() @ mat)
            cs.translate(origin)
        return cs


def CS4x4_Sagittal() -> CS4x4:
    """返回矢状位观察坐标系"""
    cs = CS4x4(axis0=[0, -1, 0], axis1=[0, 0, -1])
    cs.orthogonalize([2])
    return cs


def CS4x4_Coronal() -> CS4x4:
    """返回冠状位观察坐标系"""
    cs = CS4x4(axis0=[1, 0, 0], axis1=[0, 0, -1])
    cs.orthogonalize([2])
    return cs


def CS4x4_Axial() -> CS4x4:
    """返回横断位观察坐标系"""
    cs = CS4x4(axis0=[1, 0, 0], axis1=[0, 1, 0])
    cs.orthogonalize([2])
    return cs


def CS4x4_Vt(i: int, vt: Union[np.ndarray, List[float]], theta: float = 0) -> CS4x4:
    """
    返回与向量vt与其正交向量组成的坐标系

    :param i: 坐标轴次序
    :param vt: 向量
    :param theta: 绕向量vt的旋转角
    :return: 坐标系
    """
    axes = [np.zeros(3), np.zeros(3), np.zeros(3)]
    axes[i] = np.array(vt)
    r = np.linalg.norm(axes[i])

    if axes[i][0] ** 2 > axes[i][1] ** 2 and axes[i][0] ** 2 > axes[i][2] ** 2:
        x, y, z = 0, 1, 2
    elif axes[i][1] ** 2 > axes[i][2] ** 2:
        x, y, z = 1, 2, 0
    else:
        x, y, z = 2, 0, 1

    a, b, c = axes[i][x] / r, axes[i][y] / r, axes[i][z] / r
    tmp = (a ** 2 + c ** 2) ** 0.5

    if theta != 0:
        sint = np.sin(theta)
        cost = np.cos(theta)

        axes[(i + 1) % 3][x] = (c * cost - a * b * sint) / tmp
        axes[(i + 1) % 3][y] = sint * tmp
        axes[(i + 1) % 3][z] = (-a * cost - b * c * sint) / tmp

        axes[(i + 2) % 3][x] = (-c * sint - a * b * cost) / tmp
        axes[(i + 2) % 3][y] = cost * tmp
        axes[(i + 2) % 3][z] = (a * sint - b * c * cost) / tmp
    else:
        axes[(i + 1) % 3][x] = c / tmp
        axes[(i + 1) % 3][y] = 0
        axes[(i + 1) % 3][z] = -a / tmp

        axes[(i + 2) % 3][x] = -a * b / tmp
        axes[(i + 2) % 3][y] = tmp
        axes[(i + 2) % 3][z] = -b * c / tmp

    return CS4x4(axis0=axes[0], axis1=axes[1], axis2=axes[2])


def CS4x4_InterLandmarks(pts0: List[np.ndarray], pts1: List[np.ndarray]) -> vmi.CS4x4:
    """
    返回对应点列变换，计算点列pts0到点列pts1的最小二乘变换

    :param pts0: 点列
    :param pts1: 点列
    :return: 变换
    """
    if len(pts0) > 2 and len(pts1) > 2:
        pts = [vtk.vtkPoints(), vtk.vtkPoints()]
        for pt in pts0:
            pts[0].InsertNextPoint([pt[0], pt[1], pt[2]])
        for pt in pts1:
            pts[1].InsertNextPoint([pt[0], pt[1], pt[2]])

        lt = vtk.vtkLandmarkTransform()
        lt.SetSourceLandmarks(pts[0])
        lt.SetTargetLandmarks(pts[1])
        lt.SetModeToRigidBody()
        lt.Update()

        mat = vtk.vtkMatrix4x4()
        lt.GetMatrix(mat)

        cs = CS4x4()
        cs.setMatrix4x4(mat)
        return cs


def ptOnPlane(pt: np.ndarray, origin: np.ndarray, normal: np.ndarray) -> np.ndarray:
    """
    计算点到平面投影

    :param pt: 初始点
    :param origin: 相关平面经过一点
    :param normal: 相关平面法向量
    :return: 投影后的点
    """
    pt, origin, normal = np.array(pt), np.array(origin), np.array(normal)
    return pt - normal * np.dot(pt - origin, normal)


def vtOnPlane(vt: np.ndarray, normal: np.ndarray) -> np.ndarray:
    """
    计算向量到平面投影

    :param vt: 初始向量
    :param normal: 相关平面法向量
    :return: 投影后的向量
    """
    vt, normal = np.array(vt), np.array(normal)
    n2 = np.dot(normal, normal)
    n2 = n2 if n2 != 0 else 1
    t = np.dot(vt, normal)
    return vt - t * normal / n2


def vtOnVector(vt: np.ndarray, vector: np.ndarray) -> np.ndarray:
    """
    计算向量到向量投影
    :param vt: 初始向量
    :param vector: 相关向量
    :return: 投影后的向量
    """
    vt, vector = np.array(vt[:3]), np.array(vector[:3])
    dot = np.dot(vector, vector)
    if dot == 0:
        return np.zeros(3)
    return np.dot(vt, vector) / dot * vector


def vtAngle(vt0: Union[np.ndarray, List[float]], vt1: Union[np.ndarray, List[float]]) -> np.ndarray:
    """
    计算两个向量夹角

    :param vt0: 向量0
    :param vt1: 向量1
    :return:
    """
    vt0 = np.array(vt0)
    vt1 = np.array(vt1)
    angle = np.arccos(np.dot(vt0, vt1) / np.linalg.norm(vt0) / np.linalg.norm(vt1))
    angle = np.rad2deg(angle)
    return angle


def rayPlane(plane_origin: Union[np.ndarray, List[float]],
             plane_normal: Union[np.ndarray, List[float]],
             ray_origin: Union[np.ndarray, List[float]],
             ray_vt: Union[np.ndarray, List[float]]) -> Optional[np.ndarray]:
    """
    计算射线与平面交点

    :param plane_origin: 平面过一点
    :param plane_normal: 平面法向量
    :param ray_origin: 射线起点
    :param ray_vt: 射线方向
    :return: 交点坐标，无交点则返回None
    """
    plane_normal, plane_origin = np.array(plane_normal), np.array(plane_origin)
    ray_origin, ray_vt = np.array(ray_origin), np.array(ray_vt)

    if np.linalg.norm(plane_normal) == 0:
        raise Exception('plane_normal is {}'.format(plane_normal))
    if np.linalg.norm(ray_vt) == 0:
        raise Exception('ray_vt is {}'.format(ray_vt))

    ray_vt = ray_vt / np.linalg.norm(ray_vt)

    num = np.dot(plane_normal, plane_origin) - np.dot(plane_normal, ray_origin)
    den = np.dot(plane_normal, ray_vt)

    fabsden = -den if den < 0 else den
    fabstolerance = -num * 1e-6 if num < 0 else num * 1e-6

    if fabsden <= fabstolerance:
        return None

    x = ray_origin + num / den * ray_vt
    return x


def convert_base(number_dec: int, n: int = 36, codec: str = '0123456789abcdefghijklmnopqrstuvwxyz') -> str:
    rem_stack = []
    while number_dec > 0:
        rem = number_dec % n
        rem_stack.append(rem)
        number_dec //= n

    number_n = ""
    while len(rem_stack):
        number_n = number_n + codec[rem_stack.pop()]
    return number_n


def convert_base_inv(number_str: str, n: int = 36, codec: str = '0123456789abcdefghijklmnopqrstuvwxyz') -> int:
    number_dec = 0
    for i in range(len(number_str)):
        x = codec.find(number_str[i])
        if x >= 0:
            number_dec += x * n ** (len(number_str) - i - 1)
        else:
            raise Exception('codec not support')
    return number_dec



import setuptools
import vmi

name = vmi.__name__
version = vmi.__version__

with open('README.md', 'r', encoding='UTF-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name=name,
    version=version,
    author='sjtu_6547',
    author_email='88172828@qq.com',
    description='Visual Medical Innovation powered by WANGYiping',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://www.med-3d.top:1111/f/e386e5d3c12f4aaeb523/',
    keywords='Qt PySide2 SimpleITK itk vtk pythonocc-core OpenCASCADE CGAL',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=['PySide2>=5.14',
                      'SimpleITK>=1.2',
                      'vtk>=8.1',
                      'pydicom>=1.4',
                      'numba>=0.48',
                      'numpy>=1.18',
                      'scipy>=1.4',
                      'lz4>=3.0'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Multimedia :: Graphics :: 3D Modeling',
        'Topic :: Multimedia :: Graphics :: 3D Rendering',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)



import os
import pathlib
import pickle
import pprint
import shutil
import sys
import threading
import time
from typing import Optional, List

import PySide2
import lz4
import scipy
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from PySide2.QtWinExtras import *
from lz4.frame import compress, decompress, COMPRESSIONLEVEL_MINHC

import vmi

app = QApplication()
app.setOrganizationName('vmi')
app.setApplicationName('vmi')
app.setStyle(QStyleFactory.create('Fusion'))

desktopPath: str = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
appDataPath: str = QStandardPaths.writableLocation(QStandardPaths.AppDataLocation)
os.makedirs(appDataPath, exist_ok=True)

appViews = []

_AppWindow = None
_TaskbarButton = QWinTaskbarButton()
_Progress = _TaskbarButton.progress()


def appexit() -> None:
    """退出主程序"""
    shutil.rmtree(appDataPath, ignore_errors=True)
    sys.exit()


def appwindow(w: QWidget) -> None:
    """设置主程序窗口"""
    global _AppWindow
    _AppWindow = w
    _TaskbarButton.setParent(w)
    _TaskbarButton.setWindow(w.windowHandle())
    _Progress.setVisible(True)
    _Progress.setRange(0, 100)
    _Progress.setValue(0)


def appupdate():
    app.processEvents(QEventLoop.ExcludeUserInputEvents)


def appexec(w: QWidget, width: int = 1600, height: int = 900) -> int:
    """
    设置主程序窗口，居中显示，启动主程序

    :param w: 主程序窗口
    :param width: 窗口宽度
    :param height: 窗口高度
    :return: int - 主程序结束状态
    """
    appwindow(w)
    rect = QGuiApplication.screens()[0].geometry()
    rect = QRect(rect.center() - QPoint(int(width / 2), int(height / 2)), QSize(width, height))
    w.setGeometry(rect)
    w.showMaximized()
    return app.exec_()


def appwait(thread: threading.Thread, progress: Optional[List[int]] = None) -> None:
    """
    执行后台线程并等待，默认在windows系统任务栏循环更新进度，设置progress=[value]可以显示真实进度

    :param thread: 后台线程
    :param progress: 进度
    :return: None
    """
    while thread.is_alive():
        if progress is None:
            value = (_Progress.value() + 5) % 100
        else:
            value = progress[0]
        _Progress.setValue(value)
        time.sleep(0.1)
    _Progress.setValue(0)


def app_about(self):
    text = ''
    text += '软件名称：{}'.format(app.applicationName())
    text += '\n软件版本：{}'.format(app.applicationVersion())
    text += '\n软件环境：'
    text += '\nPython\t{}.{}.{} {} {}'.format(
        sys.version_info.major, sys.version_info.minor, sys.version_info.micro,
        sys.version_info.releaselevel, sys.version_info.serial)
    text += '\nvmi\t{}'.format(vmi.__version__)
    text += '\nPySide2\t{}'.format(PySide2.__version__)
    text += '\nSimpleITK\t{}'.format(vmi.sitk.Version.VersionString())
    text += '\npydicom\t{}'.format(vmi.pydicom.__version__)
    text += '\nvtk\t{}'.format(vmi.vtk.VTK_VERSION)
    text += '\nnumba\t{}'.format(vmi.numba.__version__)
    text += '\nnumpy\t{}'.format(vmi.np.__version__)
    text += '\nscipy\t{}'.format(scipy.__version__)
    text += '\nlz4\t{}'.format(lz4.__version__)
    vmi.askInfo(text, '关于')


class Main(QMainWindow):
    """通用应用程序外壳"""

    def __init__(self, return_globals):
        QMainWindow.__init__(self)
        brush = QBrush(QColor(255, 255, 255, 255))
        palette = QPalette()
        palette.setBrush(QPalette.Active, QPalette.Base, brush)
        palette.setBrush(QPalette.Active, QPalette.Window, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Base, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Window, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Base, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Window, brush)
        self.setPalette(palette)

        self._ScrollArea = QScrollArea()
        self._ScrollArea.setWidgetResizable(True)
        self.setCentralWidget(self._ScrollArea)

        self._ScrollAreaWidget = QWidget()
        self._Layout = QGridLayout(self._ScrollAreaWidget)
        self._ScrollArea.setWidget(self._ScrollAreaWidget)

        self._AppName = str()
        self._AppVersion = str()
        self._SaveFile = str()
        self._SaveDateTime = str()
        self.updateWindowTitle()

        menubar = QMenuBar(self)
        self.setMenuBar(menubar)

        menuFile = QMenu('文件', menubar)
        menubar.addAction(menuFile.menuAction())

        actionDataSheet = QAction('数据表', self)
        actionDataSheet.setShortcut(QKeySequence('Ctrl+D'))
        actionDataSheet.triggered.connect(self.actionDataSheet)
        menuFile.addAction(actionDataSheet)

        actionSave = QAction('保存', self)
        actionSave.setShortcut(QKeySequence('Ctrl+S'))
        actionSave.triggered.connect(self.actionSave)
        menuFile.addAction(actionSave)

        menuHelp = QMenu('帮助', menubar)
        menubar.addAction(menuHelp.menuAction())

        actionAbout = QAction('关于', self)
        actionAbout.triggered.connect(app_about)
        menuHelp.addAction(actionAbout)

        self.dataTree = QTreeWidget()
        self.dataTree.setWindowTitle('数据表')
        self.dataTree.setColumnCount(3)
        self.dataTree.setHeaderLabels(['名称', '类型', '内容'])
        self.dataTree.setAlternatingRowColors(True)
        self.dataTree.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.dataTree.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)

        statusbar = QStatusBar(self)
        self.setStatusBar(statusbar)

        self.globals = return_globals
        self.excludeKeys = list(self.globals().keys())

    def appName(self):
        return self._AppName

    def setAppName(self, name) -> None:
        """设置程序名称"""
        self._AppName = name
        app.setApplicationName(self._AppName)
        self.updateWindowTitle()

    def setAppVersion(self, version) -> None:
        """设置程序版本"""
        self._AppVersion = version
        app.setApplicationVersion(self._AppVersion)
        self.updateWindowTitle()

    def setSaveFile(self, save_file) -> None:
        """设置程序存档文件"""
        self._SaveFile = str(save_file)
        self.updateWindowTitle()

    def setSaveDateTime(self, save_date_time) -> None:
        """设置程序存档文件时间"""
        self._SaveDateTime = str(save_date_time)
        self.updateWindowTitle()

    def layout(self) -> QGridLayout:
        """返回布局器"""
        return self._Layout

    def scrollArea(self) -> QScrollArea:
        return self._ScrollArea

    def closeEvent(self, ev: QCloseEvent):
        if vmi.askYesNo('退出？'):
            ev.accept()
            appexit()
        else:
            ev.ignore()

    def actionDataSheet(self):
        self.dataTree.clear()
        objs = self.globals()
        objs = {kw: objs[kw] for kw in objs if kw not in self.excludeKeys}
        for kw in objs:
            content = pprint.pformat(objs[kw], width=500, compact=True)
            item = QTreeWidgetItem([kw, repr(type(objs[kw])), content])
            self.dataTree.addTopLevelItem(item)
        for i in range(self.dataTree.columnCount()):
            self.dataTree.resizeColumnToContents(i)
        self.dataTree.showMaximized()

    def actionSave(self):
        f = vmi.askSaveFile(nameFilter='*.vmi', title='存档')
        if f is not None:
            self._SaveFile = f
            self._SaveDateTime = time.strftime("%Y-%m-%d %H-%M-%S", time.localtime())

            ba = self.dumps()
            pathlib.Path(f).write_bytes(ba)
            vmi.askInfo('保存完成 {:.2f} MB'.format(len(ba) / (2 ** 20)))
            self.updateWindowTitle()

    def dumps(self) -> bytes:
        """存档当前程序状态为字节码"""
        objs = self.globals()
        objs = {kw: objs[kw] for kw in objs if kw not in self.excludeKeys}
        for kw in ['_AppName', '_AppVersion', '_SaveFile', '_SaveDateTime']:
            if hasattr(self, kw):
                objs[kw] = getattr(self, kw)
        ba = pickle.dumps(objs)
        ba = compress(ba, compression_level=COMPRESSIONLEVEL_MINHC)
        return ba

    def loads(self, ba: bytes):
        """读取存档字节码，更新主文件全局变量"""
        ba = decompress(ba)
        objs = pickle.loads(ba)
        for kw in ['_AppName', '_AppVersion', '_SaveFile', '_SaveDateTime']:
            if hasattr(self, kw):
                getattr(self, kw.replace('_', 'set', 1))(objs[kw])
                del objs[kw]
        self.updateWindowTitle()
        self.globals().update(objs)

    def screenWidth(self) -> int:
        """返回屏幕宽度，单位是像素"""
        return QGuiApplication.screens()[0].geometry().width()

    def updateWindowTitle(self):
        """更新窗体标题文字"""
        self.setWindowTitle('{} {} {} {}'.format(self._AppName, self._AppVersion, self._SaveFile, self._SaveDateTime))


_scripts = pathlib.Path(sys.executable).parent / 'Scripts'
_lupdate = str(_scripts / 'pyside2-lupdate')
_uic = str(_scripts / 'pyside2-uic')
_rcc = str(_scripts / 'pyside2-rcc')

_pyside2 = pathlib.Path(sys.executable).parent / 'Lib/site-packages/PySide2'
_designer = str(_pyside2 / 'designer')
_linguist = str(_pyside2 / 'linguist')
_lrelease = str(_pyside2 / 'lrelease')


def app_uic(ui_dir: str = 'ui', py_dir: str = '.', prefix: str = 'Ui_') -> None:
    """
    将ui_dir目录下的所有.ui文件转换到py_dir目录下.py文件并附加前缀ui_

    :param ui_dir: 界面文件（.ui）目录
    :param py_dir: 转换后的源代码文件（.py）目录
    :param prefix: 转换后的文件名前缀
    :return: None
    """
    py_dir = pathlib.Path(py_dir)

    for ui in pathlib.Path(ui_dir).rglob('*.ui'):
        if ui.is_file():
            py = py_dir / (prefix + ui.with_suffix('.py').name)
            os.system(_uic + ' ' + str(ui) + ' -o ' + str(py) + ' -x')
            print('[uic.exe]', _uic, ui, '-o', py, '-x')


def app_lupdate(py_dir: str = '.', ts_file: str = 'zh_CN.ts') -> None:
    """
    将py_dir目录下的所有.py文件更新翻译到同目录下的.ts文件ts_file

    :param py_dir: 源代码文件（.py)目录
    :param ts_file: 转换后的翻译文件（.ts）路径
    :return: None
    """
    py_dir = pathlib.Path(py_dir)
    pys = str()

    for py in pathlib.Path(py_dir).rglob('*.py'):
        if py.is_file():
            pys += str(py) + ' '

    if len(pys) > 0:
        ts_file = py_dir / ts_file
        os.system(_lupdate + ' ' + str(pys) + ' -ts ' + str(ts_file))
        print('[lupdate.exe]', _lupdate, pys, '-ts', ts_file)


def app_lrelease(ts_file: str = './zh_CN.ts',
                 qm_file: str = 'qrc/tr/zh_CN.qm') -> None:
    """
    将可读翻译文件ts_file转换为二进制翻译文件qm_file

    :param ts_file: 可读翻译文件（.ts）
    :param qm_file: 转换后的二进制翻译文件（.qm）
    :return: None
    """
    os.system(_lrelease + ' ' + ts_file + ' -qm ' + qm_file)
    print('[lrelease.exe]', _lrelease, ts_file, '-qm', qm_file)
    os.system(_linguist + ' ' + './zh_CN.ts')


def app_rcc(rc_dir: str = 'qrc', qrc_file: str = 'vmi.qrc', py_file: str = 'vrc.py') -> None:
    """
    将rc_dir目录下的所有文件转换为.qrc资源文件qrc_file，再转换为.py文件py_file

    :param rc_dir: 资源文件目录
    :param qrc_file: 资源文件（.qrc）
    :param py_file: 转换后的源代码文件（.py）
    :return: None
    """
    rc_dir = pathlib.Path(rc_dir)
    indent = ' ' * 4
    rcs = str()

    for rc in pathlib.Path(rc_dir).rglob('*'):
        if rc.is_file():
            rc = str(rc).replace('\\', '/')
            rcs += indent * 2 + '<file>' + rc + '</file>' + '\n'

    if len(rcs) > 0:
        t = '<!DOCTYPE RCC>'
        t += '<RCC version="1.0">' + '\n'
        t += indent + '<qresource prefix="/">' + '\n'
        t += rcs
        t += indent + '</qresource>' + '\n'
        t += '</RCC>'

        with open(qrc_file, 'w') as f:
            f.write(t)

        os.system(_rcc + ' ' + qrc_file + ' -o ' + py_file)
        print('[rcc.exe]', _rcc, qrc_file, '-o', py_file)


if __name__ == '__main__':
    # app_uic()  # 更新界面.ui文件
    # app_lupdate()  # 更新翻译.ts文件
    # app_lrelease()  # 更新翻译.qm文件
    app_rcc()  # 更新.qrc文件

    # os.startfile(_designer)  # 打开界面.ui文件



import datetime
from typing import List, Optional

import docx
import docx.dml.color
import docx.document
import docx.enum.dml
import docx.enum.section
import docx.enum.style
import docx.enum.table
import docx.enum.text
import docx.oxml.ns
import docx.shared
import docx.styles.styles
import docx.text
import docx.text.paragraph
import docx.section

import vmi


def docx_document(title: str, file: str = None) -> docx.document.Document:
    doccument = docx.Document(file)
    doccument.core_properties.created = datetime.datetime.utcnow()
    doccument.core_properties.modified = datetime.datetime.utcnow()
    doccument.core_properties.title = title
    doccument.core_properties.author = vmi.app.applicationName()
    doccument.core_properties.last_modified_by = vmi.app.applicationName()
    return doccument


def docx_style(document: docx.document.Document, name: str, base: str, font: str = '等线', bold: bool = False,
               size: int = 12, rgb: List[int] = (0, 0, 0), alignment: str = 'LEFT', first_line_indent: float = 0,
               line_spacing: float = 1, space_before: float = 0.5, space_after: float = 0.5, left_indent: float = 0):
    base = document.styles[base]
    style = document.styles.add_style(name, base.type)
    style.base_style = base
    style.quick_style = True

    style.font.name = font
    style.font.bold = bold
    style.font.size = docx.shared.Pt(size)
    style.font.color.rgb = docx.shared.RGBColor(*rgb)
    style.font.no_proof = True
    style.element.rPr.rFonts.set(docx.oxml.ns.qn('w:eastAsia'), style.font.name)
    alignment = getattr(docx.enum.text.WD_PARAGRAPH_ALIGNMENT, alignment)
    style.paragraph_format.alignment = alignment
    style.paragraph_format.first_line_indent = first_line_indent * style.font.size
    style.paragraph_format.line_spacing = line_spacing
    style.paragraph_format.space_before = round(space_before * style.font.size)
    style.paragraph_format.space_after = round(space_after * style.font.size)
    style.paragraph_format.left_indent = round(left_indent * style.font.size)

    return style


def docx_section_landscape_A4(section: docx.section.Section):
    section.page_width = docx.shared.Length(10692130)
    section.page_height = docx.shared.Length(7560310)
    section.header_distance = docx.shared.Length(914400)
    section.footer_distance = docx.shared.Length(914400)
    section.orientation = docx.enum.section.WD_ORIENTATION.LANDSCAPE


def docx_section_margin_middle(section: docx.section.Section):
    section.top_margin = docx.shared.Length(914400)
    section.bottom_margin = docx.shared.Length(914400)
    section.left_margin = docx.shared.Length(685800)
    section.right_margin = docx.shared.Length(685800)


def docx_section_header(section: docx.section.Section):
    section.header.paragraphs[0].paragraph_format.tab_stops.clear_all()
    section.header.paragraphs[0].paragraph_format.tab_stops.add_tab_stop(
        round(0.5 * (section.page_width - section.left_margin - section.right_margin)),
        docx.enum.text.WD_PARAGRAPH_ALIGNMENT.CENTER)
    section.header.paragraphs[0].paragraph_format.tab_stops.add_tab_stop(
        section.page_width - section.left_margin - section.right_margin,
        docx.enum.text.WD_PARAGRAPH_ALIGNMENT.RIGHT)


if __name__ == '__main__':
    d = docx_document('设计确认单')
    docx_section_landscape_A4(d.sections[0])
    docx_section_margin_middle(d.sections[0])
    docx_section_header(d.sections[0])

    yw_title = docx_style(d, '影为标题', 'Normal', size=26, alignment='CENTER')
    yw_header = docx_style(d, '影为页眉', 'Header', size=10)
    yw_body_left = docx_style(d, '影为正文', 'Body Text')
    yw_body_left_indent = docx_style(d, '影为正文首行缩进', 'Body Text', first_line_indent=2)
    yw_body_center = docx_style(d, '影为正文居中', 'Body Text', alignment='CENTER')
    yw_table = docx_style(d, '影为表格', 'Table Grid', alignment='CENTER')

    d.sections[0].header.paragraphs[0].text = '{}\t\t影为医疗科技（上海）有限公司\t\t[表单号]'.format(
        datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S"))
    d.sections[0].header.paragraphs[0].style = yw_header


    def add_row_text(table, texts, styles):
        r = len(table.add_row().table.rows) - 1
        for c in range(len(table.columns)):
            table.cell(r, c).paragraphs[0].text = texts[c]
            table.cell(r, c).paragraphs[0].style = styles[c]


    def add_row_picture(table, file):
        r = len(table.add_row().table.rows) - 1
        table.cell(r, 1).merge(table.cell(r, 3))
        table.cell(r, 1).paragraphs[0].add_run().add_picture(file, width=0.5 * table.cell(r, 1).width)


    top_level_table = d.add_table(1, 4, yw_table)
    top_level_table.autofit = True

    top_level_table.cell(0, 0).merge(top_level_table.cell(0, 3))
    top_level_table.cell(0, 0).paragraphs[0].text = '设计确认单\nDesign Confirmation'
    top_level_table.cell(0, 0).paragraphs[0].style = yw_title

    add_row_text(top_level_table, ['产品名称\nProduct name', '全髋关节置换手术导板\nTHA surgical guide',
                              '序列号\nSerial number', 'THK6KSDJFHKJ'], [yw_body_center] * 4)
    add_row_text(top_level_table, ['型号\nModel', 'TH', '规格\nSpecification', '定制式'], [yw_body_center] * 4)

    add_row_picture(top_level_table, 'c:/py/1.jpg')

    d.save('c:/py/demo.docx')



import glob
import vtk
import numpy as np
import pandas as pd

import vmi

from PySide2.QtCore import *
from PySide2.QtWidgets import *

SizePolicyMinimum = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)


def input_dir():
    global file_dir

    s = QSettings()
    title = '请选择文件夹 (Please select directory)'

    d = QFileDialog(None, title, str(s.value('Last' + title, defaultValue=vmi.desktopPath)))
    d.setAcceptMode(QFileDialog.AcceptOpen)
    d.setFileMode(QFileDialog.Directory)
    d.setViewMode(QFileDialog.Detail)
    d.setFilter(QDir.AllEntries | QDir.NoSymLinks | QDir.NoDotAndDotDot)

    if d.exec_() != QDialog.Accepted or len(d.selectedFiles()) < 1:
        return

    s.setValue('Last' + title, d.selectedFiles()[0])
    file_dir = d.selectedFiles()[0]

    input_list()


def input_list():
    global file_list, file_name, cup_origin, cup_radius, cup_cs, cup_angle

    if file_dir:
        file_list = glob.glob(r'{}/*.nii.gz'.format(file_dir))

        npy_dir = str(file_dir[:-6]) + '/npy'
        npy_list = glob.glob(r'{}/*.npy'.format(npy_dir))

        mark_dir = str(file_dir[:-6]) + '/mark/npy'
        mark_list = glob.glob(r'{}/*.npy'.format(mark_dir))

        tree = QTreeWidget()
        tree.setColumnCount(1)
        tree.setHeaderLabels(['File Names'])

        for x in file_list:
            suffix = str('\\' + x.split('\\')[1])[:-7]
            index = npy_dir + suffix + '.npy'
            mark_index = npy_dir[:-3] + 'mark/npy' + suffix + '_mark.npy'

            if index in npy_list:
                if mark_index not in mark_list:
                    item = QTreeWidgetItem()
                    item.setText(0, x)
                    tree.addTopLevelItem(item)

        dialog = QDialog()
        dialog.setSizePolicy(SizePolicyMinimum)
        dialog.setWindowTitle('选择文件')
        dialog.setLayout(QVBoxLayout())
        dialog.setMinimumWidth(500)
        dialog.layout().addWidget(tree)

        button = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
        button.setSizePolicy(SizePolicyMinimum)
        button.accepted.connect(dialog.accept)
        button.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button)

        if dialog.exec_() == QDialog.Accepted:
            file_name = tree.currentItem().text(0)

            cup_origin, cup_radius, cup_cs, cup_angle = [[0, 0, 0], [0, 0, 0]], [24, 24], [[], []], [[20, 40], [20, 40]]

            radius_boxes[0].draw_text('右侧半径 {:.1f}'.format(cup_radius[0]))
            radius_boxes[1].draw_text('左侧半径 {:.1f}'.format(cup_radius[1]))
            angle_boxes[0].draw_text('右侧前倾 {:.1f}'.format(cup_angle[0][0]))
            angle_boxes[1].draw_text('右侧外展 {:.1f}'.format(cup_angle[0][1]))
            angle_boxes[2].draw_text('左侧前倾 {:.1f}'.format(cup_angle[1][0]))
            angle_boxes[3].draw_text('左侧外展 {:.1f}'.format(cup_angle[1][1]))

            init_pcs()
            input_image()


def check_list():
    global file_list, file_name, cup_origin, cup_radius, cup_angle

    if file_dir:
        file_list = glob.glob(r'{}/*.nii.gz'.format(file_dir))

        npy_dir = str(file_dir[:-6]) + '/npy'
        npy_list = glob.glob(r'{}/*.npy'.format(npy_dir))

        mark_dir = str(file_dir[:-6]) + '/mark/npy'
        mark_list = glob.glob(r'{}/*.npy'.format(mark_dir))

        tree = QTreeWidget()
        tree.setColumnCount(1)
        tree.setHeaderLabels(['File Names'])

        for x in file_list:
            suffix = str('\\' + x.split('\\')[1])[:-7]
            index = npy_dir + suffix + '.npy'
            mark_index = npy_dir[:-3] + 'mark/npy' + suffix + '_mark.npy'

            if index in npy_list:
                if mark_index in mark_list:
                    item = QTreeWidgetItem()
                    item.setText(0, x)
                    tree.addTopLevelItem(item)

        dialog = QDialog()
        dialog.setSizePolicy(SizePolicyMinimum)
        dialog.setWindowTitle('选择文件')
        dialog.setLayout(QVBoxLayout())
        dialog.setMinimumWidth(500)
        dialog.layout().addWidget(tree)

        button = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
        button.setSizePolicy(SizePolicyMinimum)
        button.accepted.connect(dialog.accept)
        button.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button)

        if dialog.exec_() == QDialog.Accepted:
            file_name = tree.currentItem().text(0)

            init_pcs()
            input_image()

            npy_name = str(file_name).split('image')[0] + 'mark/npy'
            suffix = str(file_name).split('image')[1][:-7]

            mark_dict = np.load(npy_name + suffix + '_mark.npy', allow_pickle=True).item()

            cup_origin = [mark_dict['右侧臼杯中心'], mark_dict['左侧臼杯中心']]
            cup_radius = [mark_dict['右侧臼杯半径'], mark_dict['左侧臼杯半径']]

            if '右侧前倾角' in mark_dict.keys():
                cup_angle = [[mark_dict['右侧前倾角'], mark_dict['右侧外展角']],
                             [mark_dict['左侧前倾角'], mark_dict['左侧外展角']]]

            update_slice_view()
            update_cup()

            radius_boxes[0].draw_text('右侧半径 {:.1f}'.format(cup_radius[0]))
            radius_boxes[1].draw_text('左侧半径 {:.1f}'.format(cup_radius[1]))
            angle_boxes[0].draw_text('右侧前倾 {:.1f}'.format(cup_angle[0][0]))
            angle_boxes[1].draw_text('右侧外展 {:.1f}'.format(cup_angle[0][1]))
            angle_boxes[2].draw_text('左侧前倾 {:.1f}'.format(cup_angle[1][0]))
            angle_boxes[3].draw_text('左侧外展 {:.1f}'.format(cup_angle[1][1]))


def input_image():
    global series_image, cup_origin

    if file_name:
        reader = vtk.vtkNIFTIImageReader()
        reader.SetFileName(file_name)
        reader.Update()

        series_image = reader.GetOutput()
        series_volume.setData(series_image)
        series_view.setCamera_FitAll()
        series_view.updateAtOnce()

        # 初始化臼杯中心
        x_offset = 0.25 * (vmi.imBounds(series_image)[1] - vmi.imBounds(series_image)[0])
        y_offset = 0.25 * (vmi.imBounds(series_image)[3] - vmi.imBounds(series_image)[2])
        cup_origin[0] = vmi.imCenter(series_image) - x_offset * pelvis_cs.axis(0) + y_offset * pelvis_cs.axis(2)
        cup_origin[1] = vmi.imCenter(series_image) + x_offset * pelvis_cs.axis(0) + y_offset * pelvis_cs.axis(2)

        init_slice_view()
        update_cup()

        for v in right_views + left_views:
            v.setCamera_FitAll()

        vmi.appupdate()


def init_pcs():
    global pcs, pelvis_cs

    if file_name:
        npy_name = str(file_name).split('image')[0] + 'npy'
        suffix = str(file_name).split('image')[1][:-7]

        pt_dict = np.load(npy_name + suffix + '.npy', allow_pickle=True).item()

        # # 验证标志点
        # test_prop = vmi.PolyActor(series_view, color=[0.4, 0.6, 1], line_width=3, pickable=True)
        # test_prop.setData(vmi.pdAppend([vmi.pdSphere(2.5, pt_dict['髂前上棘'][0]),
        #                                 vmi.pdSphere(2.5, pt_dict['髂前上棘'][1]),
        #                                 vmi.pdSphere(2.5, pt_dict['耻骨结节'][0]),
        #                                 vmi.pdSphere(2.5, pt_dict['耻骨结节'][1]),
        #                                 vmi.pdSphere(2.5, pt_dict['髂翼最高点'][0]),
        #                                 vmi.pdSphere(2.5, pt_dict['髂翼最高点'][1]),
        #                                 vmi.pdSphere(2.5, pt_dict['坐骨结节'][0]),
        #                                 vmi.pdSphere(2.5, pt_dict['坐骨结节'][1])]))

        pcs = {'x': [1, 0, 0],
               'y': [0, 1, 0],
               'z': [0, 0, 1]}

        # APP
        app_pts = vtk.vtkPoints()
        app_pts.InsertNextPoint(pt_dict['髂前上棘'][0])
        app_pts.InsertNextPoint(pt_dict['髂前上棘'][1])
        app_pts.InsertNextPoint(pt_dict['耻骨结节'][0])
        app_pts.InsertNextPoint(pt_dict['耻骨结节'][1])

        pcs['y'] = -vmi.pdOBB_Pts(app_pts)['axis'][2]

        # MPP
        mpp_pts = [[pt_dict['髂前上棘'][0], pt_dict['髂前上棘'][1]],
                   [pt_dict['髂后上棘'][0], pt_dict['髂后上棘'][1]],
                   [pt_dict['坐骨棘'][0], pt_dict['坐骨棘'][1]],
                   [pt_dict['泪滴'][0], pt_dict['泪滴'][1]]]

        for j in range(4):
            if mpp_pts[j][0][0] > mpp_pts[j][1][0]:
                n = mpp_pts[j][1][0]
                mpp_pts[j][1][0] = mpp_pts[j][0][0]
                mpp_pts[j][0][0] = n

        mpp_vec = [(mpp_pts[0][1] - mpp_pts[0][0]) / vtk.vtkMath.Normalize(mpp_pts[0][1] - mpp_pts[0][0]),
                   (mpp_pts[1][1] - mpp_pts[1][0]) / vtk.vtkMath.Normalize(mpp_pts[1][1] - mpp_pts[1][0]),
                   (mpp_pts[2][1] - mpp_pts[2][0]) / vtk.vtkMath.Normalize(mpp_pts[2][1] - mpp_pts[2][0]),
                   (mpp_pts[3][1] - mpp_pts[3][0]) / vtk.vtkMath.Normalize(mpp_pts[3][1] - mpp_pts[3][0])]

        pcs['x'] = (mpp_vec[0] + mpp_vec[1] + mpp_vec[2] + mpp_vec[3]) / vtk.vtkMath.Normalize(
            mpp_vec[0] + mpp_vec[1] + mpp_vec[2] + mpp_vec[3])
        pcs['x'] = vmi.vtOnPlane(pcs['x'], pcs['y'])

        vtk.vtkMath.Cross(pcs['x'], pcs['y'], pcs['z'])

        if vtk.vtkMath.DegreesFromRadians(
                vtk.vtkMath.AngleBetweenVectors(pcs['x'], [1, 0, 0])) > 90:
            pcs['x'] = [-pcs['x'][0], -pcs['x'][1], -pcs['x'][2]]

        if vtk.vtkMath.DegreesFromRadians(
                vtk.vtkMath.AngleBetweenVectors(pcs['y'], [0, 1, 0])) > 90:
            pcs['y'] = [-pcs['y'][0], -pcs['y'][1], -pcs['y'][2]]

        if vtk.vtkMath.DegreesFromRadians(
                vtk.vtkMath.AngleBetweenVectors(pcs['z'], [0, 0, 1])) > 90:
            pcs['z'] = [-pcs['z'][0], -pcs['z'][1], -pcs['z'][2]]

        pelvis_cs = vmi.CS4x4(axis0=np.array(pcs['x']),
                              axis1=np.array(pcs['y']),
                              axis2=np.array(pcs['z']),
                              origin=[0, 0, 0])


def output_data():
    global file_name, series_image

    if series_image is not None:
        if file_name:
            save_name = str(file_name).split('image')[0] + 'mark/'
            suffix = str(file_name).split('image')[1][:-7]

            origin = vmi.imOrigin(series_image)
            spacing = vmi.imSpacing(series_image)
            extent = vmi.imExtent(series_image)

            stencil = [vmi.imStencil_PolyData(cup_prop[0].data(), origin, spacing, extent),
                       vmi.imStencil_PolyData(cup_prop[1].data(), origin, spacing, extent)]

            image = np.zeros(vmi.imArray_VTK(stencil[0]).shape).astype('uint8')
            image[vmi.imArray_VTK(stencil[0]) != 0] = 1
            image[vmi.imArray_VTK(stencil[1]) != 0] = 2

            vmi.imSaveFile_NIFTI(vmi.imVTK_Array(image, origin, spacing),
                                 save_name + '/image' + suffix + '_image.nii.gz')

            cup_data = {
                '右侧臼杯中心': cup_origin[0],
                '左侧臼杯中心': cup_origin[1],
                '右侧臼杯半径': cup_radius[0],
                '左侧臼杯半径': cup_radius[1],
                '右侧前倾角': cup_angle[0][0],
                '右侧外展角': cup_angle[0][1],
                '左侧前倾角': cup_angle[1][0],
                '左侧外展角': cup_angle[1][1]
            }

            np.save(save_name + '/npy' + suffix + '_mark.npy', cup_data)
            vmi.askInfo('标注文件已保存')


def output_parameter():
    if file_dir:
        file_list = glob.glob(r'{}/*.nii.gz'.format(file_dir))

        npy_dir = str(file_dir[:-6]) + '/npy'
        npy_list = glob.glob(r'{}/*.npy'.format(npy_dir))

        mark_dir = str(file_dir[:-6]) + '/mark/npy'
        mark_list = glob.glob(r'{}/*.npy'.format(mark_dir))

        tree = QTreeWidget()
        tree.setColumnCount(1)
        tree.setHeaderLabels(['File Names'])

        data = [[], [], [], [], [], []]
        name_list = []
        dis_list = [[], [], [], [], [], []]

        for x in file_list:
            suffix = str('\\' + x.split('\\')[1])[:-7]
            index = npy_dir + suffix + '.npy'
            mark_index = npy_dir[:-3] + 'mark/npy' + suffix + '_mark.npy'
            pt_index = npy_dir[:-3] + 'npy' + suffix + '.npy'

            if index in npy_list:
                if mark_index in mark_list:
                    mark_dict = np.load(mark_index, allow_pickle=True).item()
                    pt_dict = np.load(pt_index, allow_pickle=True).item()

                    # 遍历骨盆坐标系
                    pcs = {'x': [1, 0, 0],
                           'y': [0, 1, 0],
                           'z': [0, 0, 1]}

                    # APP
                    app_pts = vtk.vtkPoints()
                    app_pts.InsertNextPoint(pt_dict['髂前上棘'][0])
                    app_pts.InsertNextPoint(pt_dict['髂前上棘'][1])
                    app_pts.InsertNextPoint(pt_dict['耻骨结节'][0])
                    app_pts.InsertNextPoint(pt_dict['耻骨结节'][1])

                    pcs['y'] = -vmi.pdOBB_Pts(app_pts)['axis'][2]

                    # MPP
                    mpp_pts = [[pt_dict['髂前上棘'][0], pt_dict['髂前上棘'][1]],
                               [pt_dict['髂后上棘'][0], pt_dict['髂后上棘'][1]],
                               [pt_dict['坐骨棘'][0], pt_dict['坐骨棘'][1]],
                               [pt_dict['泪滴'][0], pt_dict['泪滴'][1]]]

                    for j in range(4):
                        if mpp_pts[j][0][0] > mpp_pts[j][1][0]:
                            n = mpp_pts[j][1][0]
                            mpp_pts[j][1][0] = mpp_pts[j][0][0]
                            mpp_pts[j][0][0] = n

                    mpp_vec = [(mpp_pts[0][1] - mpp_pts[0][0]) / vtk.vtkMath.Normalize(mpp_pts[0][1] - mpp_pts[0][0]),
                               (mpp_pts[1][1] - mpp_pts[1][0]) / vtk.vtkMath.Normalize(mpp_pts[1][1] - mpp_pts[1][0]),
                               (mpp_pts[2][1] - mpp_pts[2][0]) / vtk.vtkMath.Normalize(mpp_pts[2][1] - mpp_pts[2][0]),
                               (mpp_pts[3][1] - mpp_pts[3][0]) / vtk.vtkMath.Normalize(mpp_pts[3][1] - mpp_pts[3][0])]

                    pcs['x'] = (mpp_vec[0] + mpp_vec[1] + mpp_vec[2] + mpp_vec[3]) / vtk.vtkMath.Normalize(
                        mpp_vec[0] + mpp_vec[1] + mpp_vec[2] + mpp_vec[3])
                    pcs['x'] = vmi.vtOnPlane(pcs['x'], pcs['y'])

                    vtk.vtkMath.Cross(pcs['x'], pcs['y'], pcs['z'])

                    # 计算距离
                    dis = {'右侧臼杯中心-APP':
                               (vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['y'], pt_dict['耻骨结节'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['y'], pt_dict['耻骨结节'][1]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['y'], pt_dict['髂前上棘'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['y'], pt_dict['髂前上棘'][1])) / 4,
                           '右侧臼杯中心-MPP':
                               (vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['x'], pt_dict['耻骨结节'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['x'], pt_dict['耻骨结节'][1])) / 2,
                           '右侧臼杯中心-泪滴':
                               (vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['z'], pt_dict['泪滴'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['右侧臼杯中心'], pcs['z'], pt_dict['泪滴'][1])) / 2,
                           '左侧臼杯中心-APP':
                               (vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['y'], pt_dict['耻骨结节'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['y'], pt_dict['耻骨结节'][1]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['y'], pt_dict['髂前上棘'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['y'], pt_dict['髂前上棘'][1])) / 4,
                           '左侧臼杯中心-MPP':
                               (vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['x'], pt_dict['耻骨结节'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['x'], pt_dict['耻骨结节'][1])) / 2,
                           '左侧臼杯中心-泪滴':
                               (vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['z'], pt_dict['泪滴'][0]) +
                                vtk.vtkPlane.DistanceToPlane(mark_dict['左侧臼杯中心'], pcs['z'], pt_dict['泪滴'][1])) / 2}

                    # 添加数据
                    name_list.append(mark_index.split('/npy\\')[1].split('_')[0])
                    data[0].append(mark_dict['右侧臼杯半径'])
                    data[1].append(mark_dict['左侧臼杯半径'])
                    data[2].append(mark_dict['右侧前倾角'])
                    data[3].append(mark_dict['右侧外展角'])
                    data[4].append(mark_dict['左侧前倾角'])
                    data[5].append(mark_dict['左侧外展角'])

                    dis_list[0].append(dis['右侧臼杯中心-APP'])
                    dis_list[1].append(dis['右侧臼杯中心-MPP'])
                    dis_list[2].append(dis['右侧臼杯中心-泪滴'])
                    dis_list[3].append(dis['左侧臼杯中心-APP'])
                    dis_list[4].append(dis['左侧臼杯中心-MPP'])
                    dis_list[5].append(dis['左侧臼杯中心-泪滴'])

        data_df = pd.DataFrame({'患者姓名': name_list,
                                '右侧臼杯半径': data[0],
                                '左侧臼杯半径': data[1],
                                '右侧前倾角': data[2],
                                '右侧外展角': data[3],
                                '左侧前倾角': data[4],
                                '左侧外展角': data[5],
                                '右侧臼杯中心-APP': dis_list[0],
                                '右侧臼杯中心-MPP': dis_list[1],
                                '右侧臼杯中心-泪滴': dis_list[2],
                                '左侧臼杯中心-APP': dis_list[3],
                                '左侧臼杯中心-MPP': dis_list[4],
                                '左侧臼杯中心-泪滴': dis_list[5]})

        file = vmi.askSaveFile('xlsx', '*.xlsx', '导出 (Export) 数据')
        if file is None:
            return
        writer = pd.ExcelWriter(file)
        data_df.to_excel(writer)
        writer.save()


def update_series_opacity():
    series_volume.setOpacityScalar({series_opacity_range[0] - 1: 0,
                                    series_opacity_range[0]: 1,
                                    series_opacity_range[1]: 1,
                                    series_opacity_range[1] + 1: 0})
    series_volume.setColor({series_opacity_range[0]: [1, 1, 0.9],
                            series_opacity_range[1]: [1, 1, 0.1]})


def init_slice_view():
    if series_image is not None:
        right_slices[0].setSlicePlaneNormal(pcs['y'])
        right_slices[1].setSlicePlaneNormal(pcs['x'])
        right_slices[2].setSlicePlaneNormal(pcs['z'])
        left_slices[0].setSlicePlaneNormal(pcs['y'])
        left_slices[1].setSlicePlaneNormal(pcs['x'])
        left_slices[2].setSlicePlaneNormal(pcs['z'])

        o = np.array([0, 0, 0])
        coronal_cs = vmi.CS4x4(axis0=np.array(pcs['x']),
                               axis1=-np.array(pcs['z']),
                               axis2=np.array(pcs['y']),
                               origin=o)
        axial_cs = vmi.CS4x4(axis0=np.array(pcs['x']),
                             axis1=np.array(pcs['y']),
                             axis2=np.array(pcs['z']),
                             origin=o)

        sagittal_cs = [vmi.CS4x4(axis0=-np.array(pcs['y']),
                                 axis1=-np.array(pcs['z']),
                                 axis2=np.array(pcs['x']),
                                 origin=o),
                       vmi.CS4x4(axis0=np.array(pcs['y']),
                                 axis1=-np.array(pcs['z']),
                                 axis2=-np.array(pcs['x']),
                                 origin=o)]

        right_views[0].setCamera_FPlane(coronal_cs, 1)
        right_views[1].setCamera_FPlane(sagittal_cs[0], 1)
        right_views[2].setCamera_FPlane(axial_cs, 1)

        left_views[0].setCamera_FPlane(coronal_cs, 1)
        left_views[1].setCamera_FPlane(sagittal_cs[1], 1)
        left_views[2].setCamera_FPlane(axial_cs, 1)

        for s in right_slices:
            s.setSlicePlaneOrigin(cup_origin[0])
            s.setData(series_image)

        for s in left_slices:
            s.setSlicePlaneOrigin(cup_origin[1])
            s.setData(series_image)


def update_slice_view():
    if series_image is not None:
        for s in right_slices:
            s.setSlicePlaneOrigin(cup_origin[0])

        for s in left_slices:
            s.setSlicePlaneOrigin(cup_origin[1])


def update_cup():
    global cup_cs
    cup_cs[0] = pelvis_cs.rotate(cup_angle[0][0], pelvis_cs.axis(0))
    cup_cs[0] = cup_cs[0].rotate(cup_angle[0][1] - 180, pelvis_cs.axis(1))
    cup_cs[1] = pelvis_cs.rotate(cup_angle[1][0], pelvis_cs.axis(0))
    cup_cs[1] = cup_cs[1].rotate(180 - cup_angle[1][1], pelvis_cs.axis(1))

    cup_cs[0].setOrigin(cup_origin[0])
    cup_cs[1].setOrigin(cup_origin[1])

    right_plane = vtk.vtkPlane()
    right_plane.SetNormal(cup_cs[0].axis(2))
    right_plane.SetOrigin(cup_cs[0].origin())

    left_plane = vtk.vtkPlane()
    left_plane.SetNormal(cup_cs[1].axis(2))
    left_plane.SetOrigin(cup_cs[1].origin())

    right_pd = vmi.pdSphere(cup_radius[0], cup_cs[0].origin())
    right_pd = vmi.pdClip_Implicit(right_pd, right_plane)[1]

    left_pd = vmi.pdSphere(cup_radius[1], cup_cs[1].origin())
    left_pd = vmi.pdClip_Implicit(left_pd, left_plane)[1]

    right_pd_border = vmi.pdPolyline_Regular(cup_radius[0], cup_cs[0].origin(), cup_cs[0].axis(2))
    right_pd = vmi.pdAppend([right_pd, right_pd_border])

    left_pd_border = vmi.pdPolyline_Regular(cup_radius[1], cup_cs[1].origin(), cup_cs[1].axis(2))
    left_pd = vmi.pdAppend([left_pd, left_pd_border])

    cup_prop[0].setData(right_pd)
    cup_prop[1].setData(left_pd)

    r = 2
    pd = [vmi.pdAppend([vmi.pdCut_Implicit(cup_prop[0].data(), right_slices[0].slicePlane()),
                        vmi.pdSphere(r, cup_cs[0].origin())]),
          vmi.pdAppend([vmi.pdCut_Implicit(cup_prop[0].data(), right_slices[1].slicePlane()),
                        vmi.pdSphere(r, cup_cs[0].origin())]),
          vmi.pdAppend([vmi.pdCut_Implicit(cup_prop[0].data(), right_slices[2].slicePlane()),
                        vmi.pdSphere(r, cup_cs[0].origin())]),
          vmi.pdAppend([vmi.pdCut_Implicit(cup_prop[1].data(), left_slices[0].slicePlane()),
                        vmi.pdSphere(r, cup_cs[1].origin())]),
          vmi.pdAppend([vmi.pdCut_Implicit(cup_prop[1].data(), left_slices[1].slicePlane()),
                        vmi.pdSphere(r, cup_cs[1].origin())]),
          vmi.pdAppend([vmi.pdCut_Implicit(cup_prop[1].data(), left_slices[2].slicePlane()),
                        vmi.pdSphere(r, cup_cs[1].origin())])]
    for r in range(6):
        cup_section_prop[r].setData(pd[r])


def LeftButtonPressMove(**kwargs):
    global cup_origin
    if kwargs['picked'] in [cup_prop[0], cup_section_prop[0], cup_section_prop[1], cup_section_prop[2]]:
        cup_origin[0] += kwargs['picked'].view().pickVt_FocalPlane()
        update_slice_view()
        update_cup()
        return True

    elif kwargs['picked'] in [cup_prop[1], cup_section_prop[3], cup_section_prop[4], cup_section_prop[5]]:
        cup_origin[1] += kwargs['picked'].view().pickVt_FocalPlane()
        update_slice_view()
        update_cup()
        return True


def LeftButtonPressRelease(**kwargs):
    global series_opacity_range, cup_radius
    if kwargs['picked'] is series_opacity_range_box[0]:
        v = vmi.askInt(-10000, series_opacity_range[0], series_opacity_range[1])
        if v is not None:
            series_opacity_range[0] = v
        series_opacity_range_box[0].draw_text('min {:.0f}'.format(series_opacity_range[0]))
        update_series_opacity()

    elif kwargs['picked'] is series_opacity_range_box[1]:
        v = vmi.askInt(series_opacity_range[0], series_opacity_range[0], 10000)
        if v is not None:
            series_opacity_range[1] = v
        series_opacity_range_box[1].draw_text('min {:.0f}'.format(series_opacity_range[1]))
        update_series_opacity()

    elif kwargs['picked'] is radius_boxes[0]:
        v = vmi.askInt(0, cup_radius[0], 90, 1)
        if v is not None:
            cup_radius[0] = v
        radius_boxes[0].draw_text('右侧半径 {:.1f}'.format(cup_radius[0]))
        update_cup()

    elif kwargs['picked'] is radius_boxes[1]:
        v = vmi.askInt(0, cup_radius[1], 90, 1)
        if v is not None:
            cup_radius[1] = v
        radius_boxes[1].draw_text('左侧半径 {:.1f}'.format(cup_radius[1]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[0]:
        v = vmi.askFloat(0, cup_angle[0][0], 180, 1)
        if v is not None:
            cup_angle[0][0] = v
        angle_boxes[0].draw_text('右侧前倾 {:.1f}'.format(cup_angle[0][0]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[1]:
        v = vmi.askFloat(0, cup_angle[0][1], 180, 1)
        if v is not None:
            cup_angle[0][1] = v
        angle_boxes[1].draw_text('右侧外展 {:.1f}'.format(cup_angle[0][1]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[2]:
        v = vmi.askFloat(0, cup_angle[1][0], 180, 1)
        if v is not None:
            cup_angle[1][0] = v
        angle_boxes[2].draw_text('左侧前倾 {:.1f}'.format(cup_angle[1][0]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[3]:
        v = vmi.askFloat(0, cup_angle[1][1], 180, 1)
        if v is not None:
            cup_angle[1][1] = v
        angle_boxes[3].draw_text('左侧外展 {:.1f}'.format(cup_angle[1][1]))
        update_cup()


def NoButtonWheel(**kwargs):
    global series_opacity_range, cup_radius
    if kwargs['picked'] is series_opacity_range_box[0]:
        series_opacity_range[0] = min(max(series_opacity_range[0] + kwargs['delta'],
                                          -10000), series_opacity_range[1])
        series_opacity_range_box[0].draw_text('min {:.0f}'.format(series_opacity_range[0]))
        update_series_opacity()

    elif kwargs['picked'] is series_opacity_range_box[1]:
        series_opacity_range[1] = min(max(series_opacity_range[1] + kwargs['delta'],
                                          series_opacity_range[0]), 10000)
        series_opacity_range_box[1].draw_text('max {:.0f}'.format(series_opacity_range[1]))
        update_series_opacity()

    elif kwargs['picked'] is radius_boxes[0]:
        cup_radius[0] = min(max(cup_radius[0] + 1 * kwargs['delta'],
                                0), 90)
        radius_boxes[0].draw_text('右侧半径 {:.1f}'.format(cup_radius[0]))
        update_cup()

    elif kwargs['picked'] is radius_boxes[1]:
        cup_radius[1] = min(max(cup_radius[1] + 1 * kwargs['delta'],
                                0), 90)
        radius_boxes[1].draw_text('左侧半径 {:.1f}'.format(cup_radius[1]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[0]:
        cup_angle[0][0] = min(max(cup_angle[0][0] + 0.5 * kwargs['delta'],
                                  0), 180)
        angle_boxes[0].draw_text('右侧前倾 {:.1f}'.format(cup_angle[0][0]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[1]:
        cup_angle[0][1] = min(max(cup_angle[0][1] + 0.5 * kwargs['delta'],
                                  0), 180)
        angle_boxes[1].draw_text('右侧外展 {:.1f}'.format(cup_angle[0][1]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[2]:
        cup_angle[1][0] = min(max(cup_angle[1][0] + 0.5 * kwargs['delta'],
                                  0), 180)
        angle_boxes[2].draw_text('左侧前倾 {:.1f}'.format(cup_angle[1][0]))
        update_cup()

    elif kwargs['picked'] is angle_boxes[3]:
        cup_angle[1][1] = min(max(cup_angle[1][1] + 0.5 * kwargs['delta'],
                                  0), 180)
        angle_boxes[3].draw_text('左侧外展 {:.1f}'.format(cup_angle[1][1]))
        update_cup()


def return_globals():
    return globals()


if __name__ == '__main__':
    main = vmi.Main(return_globals)
    main.setAppName('Cup_marker')
    main.setAppVersion(vmi.version)
    main.excludeKeys += ['main']

    menu_input = main.menuBar().addMenu('输入')
    menu_input.addAction('选择文件夹').triggered.connect(input_dir)
    menu_input.addAction('未标注列表').triggered.connect(input_list)
    menu_input.addAction('查看已标注').triggered.connect(check_list)

    menu_output = main.menuBar().addMenu('输出')
    menu_output.addAction('完成标注').triggered.connect(output_data)
    menu_output.addAction('导出数据').triggered.connect(output_parameter)

    file_name, file_dir, file_list = [], [], []

    # 骨盆坐标系
    pcs = {'x': [1, 0, 0],
           'y': [0, 1, 0],
           'z': [0, 0, 1]}
    vtk.vtkMath.Cross(pcs['x'], pcs['y'], pcs['z'])
    pelvis_cs = vmi.CS4x4(axis0=np.array(pcs['x']),
                          axis1=np.array(pcs['y']),
                          axis2=np.array(pcs['z']),
                          origin=[0, 0, 0])

    # 体绘制视图
    series_dict = []
    series_image = []
    series_view = vmi.View()
    series_view.setCamera_Coronal()
    series_volume = vmi.ImageVolume(series_view)

    series_opacity_range = [200, 3000]
    update_series_opacity()

    series_opacity_range_box = [vmi.TextBox(series_view, text='min {:.0f}'.format(series_opacity_range[0]),
                                            size=[0.2, 0.06], pos=[0, 0.03], anchor=[0, 0], pickable=True),
                                vmi.TextBox(series_view, text='max {:.0f}'.format(series_opacity_range[1]),
                                            size=[0.2, 0.06], pos=[0.2, 0.03], anchor=[0, 0], pickable=True)]

    for box in series_opacity_range_box:
        box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

    # 二维视图
    right_views = [vmi.View(), vmi.View(), vmi.View()]
    left_views = [vmi.View(), vmi.View(), vmi.View()]

    for i in right_views + left_views:
        i.mouse['LeftButton']['PressMove'] = [i.mouseBlock]
        i.mouse['NoButton']['Wheel'] = [i.mouseBlock]

    right_boxes = [vmi.TextBox(right_views[i], text=['冠状位', '矢状位', '横断位'][i],
                               size=[0.2, 0.06], pos=[0, 0.03], anchor=[0, 0]) for i in range(3)]
    left_boxes = [vmi.TextBox(left_views[i], text=['冠状位', '矢状位', '横断位'][i],
                              size=[0.2, 0.06], pos=[0, 0.03], anchor=[0, 0]) for i in range(3)]

    right_slices = [vmi.ImageSlice(v) for v in right_views]
    left_slices = [vmi.ImageSlice(v) for v in left_views]

    for i in right_slices + left_slices:
        i.mouse['NoButton']['Wheel'] = [i.mouseBlock]
        i.setColorWindow_Bone()

    # 臼杯参数
    cup_origin, cup_radius, cup_cs, cup_angle = [[0, 0, 0], [0, 0, 0]], [24, 24], [[], []], [[20, 40], [20, 40]]

    box_size = [0.2, 0.06]
    radius_boxes = [vmi.TextBox(series_view, text='右侧半径 {:.1f}'.format(cup_radius[0]),
                                size=box_size, pos=[0, 0.09], anchor=[0, 0], pickable=True),
                    vmi.TextBox(series_view, text='左侧半径 {:.1f}'.format(cup_radius[1]),
                                size=box_size, pos=[0.2, 0.09], anchor=[0, 0], pickable=True)]

    angle_boxes = [vmi.TextBox(series_view, text='右侧前倾 {:.1f}'.format(cup_angle[0][0]),
                               size=box_size, pos=[0, 0.15], anchor=[0, 0], pickable=True),
                   vmi.TextBox(series_view, text='右侧外展 {:.1f}'.format(cup_angle[0][1]),
                               size=box_size, pos=[0, 0.21], anchor=[0, 0], pickable=True),
                   vmi.TextBox(series_view, text='左侧前倾 {:.1f}'.format(cup_angle[1][0]),
                               size=box_size, pos=[0.2, 0.15], anchor=[0, 0], pickable=True),
                   vmi.TextBox(series_view, text='左侧外展 {:.1f}'.format(cup_angle[1][1]),
                               size=box_size, pos=[0.2, 0.21], anchor=[0, 0], pickable=True)]

    for box in radius_boxes + angle_boxes:
        box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

    cup_prop = [vmi.PolyActor(series_view, color=[0.4, 0.6, 1], line_width=3, pickable=True),
                vmi.PolyActor(series_view, color=[0.4, 0.6, 1], line_width=3, pickable=True)]
    for p in cup_prop:
        p.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

    cup_section_prop = [vmi.PolyActor(right_views[0], color=[0.4, 0.6, 1], line_width=3, pickable=True),
                        vmi.PolyActor(right_views[1], color=[0.4, 0.6, 1], line_width=3, pickable=True),
                        vmi.PolyActor(right_views[2], color=[0.4, 0.6, 1], line_width=3, pickable=True),
                        vmi.PolyActor(left_views[0], color=[0.4, 0.6, 1], line_width=3, pickable=True),
                        vmi.PolyActor(left_views[1], color=[0.4, 0.6, 1], line_width=3, pickable=True),
                        vmi.PolyActor(left_views[2], color=[0.4, 0.6, 1], line_width=3, pickable=True)]
    for p in cup_section_prop:
        p.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

    # 视图布局
    main.layout().addWidget(series_view, 0, 2, 3, 1)
    main.layout().addWidget(right_views[0], 0, 1, 1, 1)
    main.layout().addWidget(right_views[1], 1, 1, 1, 1)
    main.layout().addWidget(right_views[2], 2, 1, 1, 1)
    main.layout().addWidget(left_views[0], 0, 3, 1, 1)
    main.layout().addWidget(left_views[1], 1, 3, 1, 1)
    main.layout().addWidget(left_views[2], 2, 3, 1, 1)

    vmi.appexec(main)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import pathlib
import pprint
import tempfile
from typing import List, Optional, Union, Dict

import numpy as np
import vtk
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

import vmi


class PolyActor(QObject, vmi.Menu, vmi.Mouse):
    """三维物体，将vtkPolyData通过vtkPolyDataMapper/vtkActor显示"""

    def __init__(self, view: vmi.View = None, visible: bool = True, pickable: bool = False, repres: str = 'surface',
                 color: Union[QColor, List[float]] = QColor('lightgray'), opacity: float = 1.0, point_size: int = 1,
                 line_width: int = 1, shade: bool = True, always_on_top: bool = False):
        """

        :param view: 视图
        :param visible: 可见
        :param pickable: 可交互
        :param repres: 表达形式 'points' / 'wireframe' / 'surface'
        :param color: 颜色
        :param opacity: 不透明度
        :param point_size: 点粗，单位是显示像素
        :param line_width: 线粗，单位是显示像素
        :param shade: 阴影
        :param always_on_top: 置顶显示，遮挡关系
        """
        QObject.__init__(self)
        vmi.Menu.__init__(self)

        vmi.Mouse.__init__(self)
        for b in self.mouse:
            for e in self.mouse[b]:
                if e == 'Enter':
                    self.mouse[b][e] = [self.mouseEnter]
                elif e == 'Leave':
                    self.mouse[b][e] = [self.mouseLeave]
                elif e == 'Press':
                    self.mouse[b][e] = [self.mousePress]
                elif e == 'PressMoveRelease':
                    self.mouse[b][e] = [self.mousePass if b == 'RightButton' else self.mouseRelease]
                elif e == 'PressRelease':
                    self.mouse[b][e] = [self.mousePass if b == 'RightButton' else self.mouseRelease]

        self._Mapper = vtk.vtkPolyDataMapper()
        self._Prop = vtk.vtkActor()
        self._Prop._Prop = self

        self._Property: vtk.vtkProperty = self._Prop.GetProperty()
        self._Prop.SetBackfaceProperty(self._Property)

        self._Data = vtk.vtkPolyData()
        self._BindData = self

        self._View = self._Visible = self._Pickable = self._Repres = None
        self._Color = self._Opacity = self._PointSize = self._LineWidth = self._Shade = self._AlwaysOnTop = None
        self.setView(view)
        self.setVisible(visible)
        self.setPickable(pickable)
        self.setRepres(repres)
        self.setColor(color)
        self.setOpacity(opacity)
        self.setPointSize(point_size)
        self.setLineWidth(line_width)
        self.setShade(shade)
        self.setAlwaysOnTop(always_on_top)

    def __setstate__(self, state):
        self.__init__()

        for kw in state['vtk']:
            vmi.vtkInstance_sets(getattr(self, kw), state['vtk'][kw])

        for kw in state['set']:
            getattr(self, kw.replace('_', 'set', 1))(state['set'][kw])

        del state['vtk']
        del state['set']
        self.__dict__.update(state)
        self.setView(self._View)

    def __getstate__(self):
        state = self.__dict__.copy()
        state = {kw: state[kw] for kw in state if kw not in ['menu', 'actions', '__METAOBJECT__']}

        state_set = {kw: getattr(self, kw) for kw in
                     ['_AlwaysOnTop', '_Color', '_Data', '_LineWidth', '_Opacity', '_Pickable', '_PointSize', '_Repres',
                      '_Shade', '_Visible']}
        state = {kw: state[kw] for kw in state if kw not in state_set}

        state_vtk = {}
        for kw in state:
            if isinstance(state[kw], vtk.vtkObject):
                if not hasattr(vmi, 'pickle_' + state[kw].__class__.__name__):
                    state_vtk[kw] = vmi.vtkInstance_gets(state[kw])
        state = {kw: state[kw] for kw in state if kw not in state_vtk}

        state['set'] = state_set
        state['vtk'] = state_vtk
        return state

    def __repr__(self):
        state = self.__getstate__()['set']
        state['_Data'] = {'GetActualMemorySize': self.data().GetActualMemorySize(),
                          'GetBounds': self.data().GetBounds(),
                          'GetCenter': self.data().GetCenter(),
                          'GetNumberOfCells': self.data().GetNumberOfCells(),
                          'GetNumberOfPoints': self.data().GetNumberOfPoints(),
                          'GetScalarRange': self.data().GetScalarRange()}
        return pprint.pformat(state, width=400, compact=True)

    def view(self) -> vmi.View:
        """返回视图"""
        return self._View

    def setView(self, view) -> None:
        """设置视图，同时更新数据关联"""
        if self._View:
            self._View.renderer().RemoveActor(self._Prop)
        self._View = view
        if self._View:
            self._View.renderer().AddActor(self._Prop)
        self.bindData()

    def updateInTime(self) -> None:
        """及时刷新视图"""
        if self._View:
            self._View.updateInTime()

    def visible(self) -> bool:
        """返回可见状态"""
        return self._Visible

    def setVisible(self, visible: bool) -> None:
        """设置可见状态"""
        self.updateInTime()
        self._Visible = visible
        self._Prop.SetVisibility(1 if visible else 0)

    def visibleToggle(self) -> None:
        """切换可见状态"""
        return self.setVisible(not self.visible())

    def pickable(self) -> bool:
        """返回可交互状态"""
        return self._Pickable

    def setPickable(self, pickable: bool) -> None:
        """设置可交互状态"""
        self.updateInTime()
        self._Pickable = pickable
        self._Prop.SetPickable(1 if self._Pickable else 0)

    def repres(self) -> str:
        """返回表达形式"""
        return self._Repres

    def setRepres(self, repres: str) -> None:
        """设置表达形式 'points' / 'wireframe' / 'surface'"""
        self.updateInTime()
        if repres == 'points':
            self._Repres = repres
            self._Property.SetRepresentationToPoints()
        elif repres == 'wireframe':
            self._Repres = repres
            self._Property.SetRepresentationToWireframe()
        elif repres == 'surface':
            self._Repres = repres
            self._Property.SetRepresentationToSurface()

    def color(self) -> QColor:
        """返回颜色"""
        return self._Color

    def setColor(self, color: Union[QColor, List[float]]) -> None:
        """设置颜色"""
        self.updateInTime()
        if isinstance(color, QColor):
            self._Color = color
        else:
            self._Color = QColor.fromRgbF(color[0], color[1], color[2])
        self._Property.SetColor(self._Color.redF(), self._Color.greenF(), self._Color.blueF())

    def opacity(self) -> float:
        """返回不透明度"""
        return self._Opacity

    def setOpacity(self, opacity: float) -> None:
        """设置不透明度"""
        self.updateInTime()
        opacity = min(max(opacity, 0.0), 1.0)
        self._Opacity = opacity
        self._Property.SetOpacity(self._Opacity)

    def pointSize(self) -> int:
        """返回点粗，单位是显示像素"""
        return self._PointSize

    def setPointSize(self, point_size: int) -> None:
        """设置点粗，单位是显示像素"""
        point_size = min(max(point_size, 1), 100)
        self._PointSize = point_size
        self._Property.SetPointSize(self._PointSize)

    def lineWidth(self) -> int:
        """返回线粗，单位是显示像素"""
        return self._LineWidth

    def setLineWidth(self, point_size: int) -> None:
        """设置线粗，单位是显示像素"""
        point_size = min(max(point_size, 1), 100)
        self._LineWidth = point_size
        self._Property.SetLineWidth(self._LineWidth)

    def shade(self) -> bool:
        """返回阴影状态"""
        return self._Shade

    def setShade(self, shade: bool) -> None:
        """设置阴影状态"""
        self.updateInTime()
        self._Shade = shade
        self._Property.SetAmbient(0 if self._Shade else 1)
        self._Property.SetDiffuse(1 if self._Shade else 0)

    def alwaysOnTop(self) -> bool:
        """返回置顶状态"""
        return self._AlwaysOnTop

    def setAlwaysOnTop(self, always_on_top: bool) -> None:
        """设置置顶状态"""
        self.updateInTime()
        self._AlwaysOnTop = always_on_top
        if self._AlwaysOnTop:
            self._Mapper.SetRelativeCoincidentTopologyLineOffsetParameters(0, -66000)
            self._Mapper.SetRelativeCoincidentTopologyPolygonOffsetParameters(0, -66000)
            self._Mapper.SetRelativeCoincidentTopologyPointOffsetParameter(-66000)
        else:
            self._Mapper.SetRelativeCoincidentTopologyLineOffsetParameters(-1, -1)
            self._Mapper.SetRelativeCoincidentTopologyPolygonOffsetParameters(-1, -1)
            self._Mapper.SetRelativeCoincidentTopologyPointOffsetParameter(-1)

    def data(self) -> vtk.vtkPolyData:
        """返回数据"""
        return self._Data

    def bindData(self, other=None) -> None:
        """关联数据到视图，可以关联其它对象，other = self 关联自身对象，other = None 关联当前对象"""
        self.updateInTime()
        if hasattr(other, '_Data'):
            self._BindData = other
        self._Mapper.SetInputData(self._BindData.data())
        self._Prop.SetMapper(self._Mapper)

    def setData(self, other) -> None:
        """设置数据，可以直接设置其它同类实例，或TopoDS_Shape / Polyhedron_3"""
        self.updateInTime()
        if hasattr(other, '_Data'):
            self.data().ShallowCopy(other.data())
        elif isinstance(other, self.data().__class__):
            self.data().ShallowCopy(other)
        elif isinstance(other, vmi.TopoDS_Shape):
            other = vmi.ccPd_Sh(other)
            self.setData(other)
        elif isinstance(other, vmi.Polyhedron_3):
            other = vmi.cgPd_Ph(other)
            self.setData(other)
        else:
            self.data().ShallowCopy(self.data().__class__())
        self.bindData()

    def scalar(self, image: Optional[vtk.vtkImageData] = None,
               lookup_table: Optional[vtk.vtkScalarsToColors] = None) -> None:
        if image and lookup_table:
            self._Mapper.SetColorModeToMapScalars()
            self._Mapper.SetUseLookupTableScalarRange(1)
            self._Mapper.SetScalarVisibility(1)
            self._Mapper.SetLookupTable(lookup_table)

            probe = vtk.vtkProbeFilter()
            probe.SetInputData(self._BindData.data())
            probe.SetSourceData(image)
            probe.Update()
            self.setData(probe.GetOutput())
        else:
            self._Mapper.SetScalarVisibility(0)
            self._Mapper.SetLookupTable(None)

    def delete(self) -> None:
        """清除视图关联，在del之前调用"""
        self.updateInTime()
        self.data().ReleaseData()
        self.setView(None)

    def mouseEnter(self, **kwargs) -> None:
        self.updateInTime()
        c = self._Color.lighter(120)
        self._Property.SetColor(c.redF(), c.greenF(), c.blueF())

    def mousePress(self, **kwargs) -> None:
        self.updateInTime()
        c = self._Color.darker(120)
        self._Property.SetColor(c.redF(), c.greenF(), c.blueF())

    def mouseRelease(self, **kwargs) -> None:
        self.updateInTime()
        self._Property.SetColor(self._Color.redF(), self._Color.greenF(), self._Color.blueF())

    def mouseLeave(self, **kwargs) -> None:
        self.updateInTime()
        self._Property.SetColor(self._Color.redF(), self._Color.greenF(), self._Color.blueF())


class ImageBox(QObject, vmi.Menu, vmi.Mouse):
    """二维图标，将QImage通过vtkImageMapper/vtkActor2D显示"""

    def __init__(self, view: vmi.View = None, visible: bool = True, pickable: bool = False,
                 image: QImage = QImage(1, 1, QImage.Format.Format_RGB32),
                 size: List[float] = (0.5, 0.5), pos: List[float] = (0.5, 0.5), anchor: List[float] = (0.5, 0.5)):
        """
        :param view: 视图
        :param visible: 可见
        :param pickable: 可交互
        :param image: 图像
        :param size: 显示大小，单位是视图宽高比例
        :param pos: 定位位置，单位是视图宽高比例
        :param anchor: 定位锚点，单位是视图宽高比例
        """
        QObject.__init__(self)
        vmi.Menu.__init__(self)

        vmi.Mouse.__init__(self)
        for b in self.mouse:
            for e in self.mouse[b]:
                if e == 'Enter':
                    self.mouse[b][e] = [self.mouseEnter]
                elif e == 'Leave':
                    self.mouse[b][e] = [self.mouseLeave]
                elif e == 'Press':
                    self.mouse[b][e] = [self.mousePress]
                elif e == 'PressMoveRelease':
                    self.mouse[b][e] = [self.mousePass if b == 'RightButton' else self.mouseRelease]
                elif e == 'PressRelease':
                    self.mouse[b][e] = [self.mousePass if b == 'RightButton' else self.mouseRelease]

        self._Mapper = vtk.vtkImageMapper()
        self._Mapper.SetColorLevel(128)
        self._Mapper.SetColorWindow(256)

        self._Prop = vtk.vtkActor2D()
        self._Prop._Prop = self
        self._Property: vtk.vtkProperty = self._Prop.GetProperty()
        self._Data = vtk.vtkImageData()
        self._BindData = self

        self._View = self._Visible = self._Pickable = None
        self._Image = QImage(1, 1, QImage.Format.Format_RGB32)
        self._Size = [0.5, 0.5]
        self._Pos = [0.5, 0.5]
        self._Anchor = [0.5, 0.5]
        self.setView(view)
        self.setVisible(visible)
        self.setPickable(pickable)
        self.draw(image=image, size=size, pos=pos, anchor=anchor)

    def __setstate__(self, state):
        self.__init__()

        for kw in state['vtk']:
            vmi.vtkInstance_sets(getattr(self, kw), state['vtk'][kw])

        for kw in state['set']:
            getattr(self, kw.replace('_', 'set', 1))(state['set'][kw])

        self.draw(image=state['draw']['_Image'], size=state['draw']['_Size'], pos=state['draw']['_Pos'],
                  anchor=state['draw']['_Anchor'])

        del state['vtk']
        del state['draw']
        del state['set']
        self.__dict__.update(state)
        self.setView(self._View)

    def __getstate__(self):
        state = self.__dict__.copy()
        state = {kw: state[kw] for kw in state if kw not in ['menu', 'actions', '__METAOBJECT__']}

        state_set = {kw: getattr(self, kw) for kw in ['_Pickable', '_Visible']}
        state = {kw: state[kw] for kw in state if kw not in state_set}

        state_draw = {kw: getattr(self, kw) for kw in ['_Anchor', '_Image', '_Pos', '_Size']}

        state_vtk = {}
        for kw in state:
            if isinstance(state[kw], vtk.vtkObject):
                if not hasattr(vmi, 'pickle_' + state[kw].__class__.__name__):
                    state_vtk[kw] = vmi.vtkInstance_gets(state[kw])
        state = {kw: state[kw] for kw in state if kw not in state_vtk}

        state['set'] = state_set
        state['draw'] = state_draw
        state['vtk'] = state_vtk
        return state

    def __repr__(self):
        state = self.__getstate__()
        state = {**state['set'], **state['draw']}
        return pprint.pformat(state, width=400, compact=True)

    def resize(self) -> None:
        self.draw()

    def data(self) -> vtk.vtkImageData:
        """返回数据"""
        return self._Data

    def bindData(self, other=None) -> None:
        """关联数据到视图，可以关联其它对象，other = self 关联自身对象，other = None 关联当前对象"""
        self.updateInTime()
        if hasattr(other, '_Data'):
            self._BindData = other
        self._Mapper.SetInputData(self._BindData.data())
        self._Prop.SetMapper(self._Mapper)

    def delete(self) -> None:
        """清除视图关联，在del之前调用"""
        self.updateInTime()
        self.data().ReleaseData()
        self.setView(None)

    def draw(self, image: QImage = None, size: List[float] = None, pos: List[float] = None,
             anchor: List[float] = None) -> None:
        """
        重绘，同时更新参数

        :param image: 图像
        :param size: 显示大小，单位是视图宽高比例
        :param pos: 定位位置，单位是视图宽高比例
        :param anchor: 定位锚点，单位是视图宽高比例
        :return: None
        """
        self.updateInTime()

        if image is not None:
            self._Image = image

        if size is not None:
            self._Size = [min(max(size[0], 0), 1), min(max(size[1], 0), 1)]

        if pos is not None:
            self._Pos = [min(max(pos[0], 0), 1), min(max(pos[1], 0), 1)]

        if anchor is not None:
            self._Anchor = [min(max(anchor[0], 0), 1), min(max(anchor[1], 0), 1)]

        if self.view():
            # 图像缩放
            size = [round(self._Size[0] * self.view().width()), round(self._Size[1] * self.view().height())]
            image = self._Image.scaled(size[0], size[1], Qt.AspectRatioMode.KeepAspectRatio)

            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.png'
                image.save(str(p), 'PNG')

                r = vtk.vtkPNGReader()
                r.SetFileName(str(p))
                r.Update()
                self.data().ShallowCopy(r.GetOutput())

            # 图像定位
            pos = [round(self._Pos[0] * self.view().width()), round(self._Pos[1] * self.view().height())]
            anchor = [round(self._Anchor[0] * image.width()), round((1 - self._Anchor[1]) * image.height())]

            self._Prop.SetPosition(pos[0] - anchor[0], self.view().height() - pos[1] - anchor[1])

    def view(self) -> vmi.View:
        """返回视图"""
        return self._View

    def setView(self, view) -> None:
        """设置视图，同时更新数据关联"""
        if self._View:
            self._View.renderer().RemoveActor(self._Prop)
        self._View = view
        if self._View:
            self._View.renderer().AddActor(self._Prop)
        self.bindData()
        self.draw()

    def updateInTime(self) -> None:
        """及时刷新视图"""
        if self._View:
            self._View.updateInTime()

    def visible(self) -> bool:
        """返回可见状态"""
        return self._Visible

    def setVisible(self, visible: bool) -> None:
        """设置可见状态"""
        self.updateInTime()
        self._Visible = visible
        self._Prop.SetVisibility(1 if visible else 0)

    def visibleToggle(self) -> None:
        """切换可见状态"""
        return self.setVisible(not self.visible())

    def pickable(self) -> bool:
        """返回可交互状态"""
        return self._Pickable

    def setPickable(self, pickable: bool) -> None:
        """设置可交互状态"""
        self.updateInTime()
        self._Pickable = pickable
        self._Prop.SetPickable(1 if self._Pickable else 0)

    def image(self) -> QImage:
        """返回图像"""
        return self._Image

    def size(self) -> List[float]:
        """显示大小，单位是视图宽高比例"""
        return self._Size

    def pos(self) -> List[float]:
        """定位位置，单位是视图宽高比例"""
        return self._Pos

    def anchor(self) -> List[float]:
        """定位锚点，单位是视图宽高比例"""
        return self._Anchor

    def mouseEnter(self, **kwargs) -> None:
        self.updateInTime()
        self._Mapper.SetColorLevel(96)
        self._Mapper.SetColorWindow(256)

    def mousePress(self, **kwargs) -> None:
        self.updateInTime()
        self._Mapper.SetColorLevel(160)
        self._Mapper.SetColorWindow(256)

    def mouseRelease(self, **kwargs) -> None:
        self.updateInTime()
        self._Mapper.SetColorLevel(128)
        self._Mapper.SetColorWindow(256)

    def mouseLeave(self, **kwargs) -> None:
        self.updateInTime()
        self._Mapper.SetColorLevel(128)
        self._Mapper.SetColorWindow(256)


class TextBox(ImageBox):
    """二维文字，将str通过vtkImageMapper/vtkActor2D显示"""

    def __init__(self, view: vmi.View = None, visible: bool = True, pickable: bool = False,
                 text: str = str(), text_font: str = '等线', text_align: str = 'AlignCenter',
                 fore_color: QColor = QColor('black'), back_color: QColor = QColor('whitesmoke'),
                 bold: bool = False, italic: bool = False, underline: bool = False,
                 size: List[float] = (0.5, 0.5), pos: List[float] = (0.5, 0.5), anchor: List[float] = (0.5, 0.5)):
        """
        :param view: 视图
        :param visible: 可见
        :param pickable: 可交互
        :param text: 文字
        :param text_font: 字体
        :param text_align: 对齐方式 'AlignCenter' / 'AlignLeft' / 'AlignRight'
        :param fore_color: 前景色
        :param back_color: 背景色
        :param bold: 粗体状态
        :param italic: 斜体状态
        :param underline: 下划线状态
        :param size: 显示大小，单位是视图宽高比例
        :param pos: 定位位置，单位是视图宽高比例
        :param anchor: 定位锚点，单位是视图宽高比例
        """
        ImageBox.__init__(self, visible=visible, size=size, pos=pos, anchor=anchor, pickable=pickable)

        self._Text = str()
        self._TextFont = '等线'
        self._TextAlign = 'AlignCenter'
        self._ForeColor = QColor('black')
        self._BackColor = QColor('whitesmoke')
        self._Bold = False
        self._Italic = False
        self._Underline = False
        self.draw_text(text=text, text_font=text_font, text_align=text_align,
                       fore_color=fore_color, back_color=back_color,
                       bold=bold, italic=italic, underline=underline,
                       size=size, pos=pos, anchor=anchor)
        self.setView(view)

    def font(self) -> str:
        """返回字体"""
        return self._TextFont

    def text(self) -> str:
        """返回文字"""
        return self._Text

    def text_align(self) -> str:
        """返回对齐方式"""
        return self._TextAlign

    def fore_color(self) -> QColor:
        """返回前景色"""
        return self._ForeColor

    def back_color(self) -> QColor:
        """返回背景色"""
        return self._BackColor

    def bold(self) -> bool:
        """返回粗体状态"""
        return self._Bold

    def italic(self) -> bool:
        """返回斜体状态"""
        return self._Italic

    def underline(self) -> bool:
        """返回下划线状态"""
        return self._Underline

    def draw_text(self, text: str = None, text_font: str = None, text_align: str = None,
                  fore_color: QColor = None, back_color: QColor = None,
                  bold: bool = None, italic: bool = None, underline: bool = None,
                  size: List[float] = None, pos: List[float] = None, anchor: List[float] = None) -> None:
        """
        重绘，同时更新参数

        :param text: 文字
        :param text_font: 字体
        :param text_align: 对齐方式 'AlignCenter' / 'AlignLeft' / 'AlignRight'
        :param fore_color: 前景色
        :param back_color: 背景色
        :param bold: 粗体状态
        :param italic: 斜体状态
        :param underline: 下划线状态
        :param size: 显示大小，单位是视图宽高比例
        :param pos: 定位位置，单位是视图宽高比例
        :param anchor: 定位锚点，单位是视图宽高比例
        :return: None
        """
        if text is not None:
            self._Text = text
        if text_font is not None:
            self._TextFont = text_font
        if text_align is not None:
            self._TextAlign = text_align
        if fore_color is not None:
            self._ForeColor = fore_color
        if back_color is not None:
            self._BackColor = back_color
        if bold is not None:
            self._Bold = bold
        if italic is not None:
            self._Italic = italic
        if underline is not None:
            self._Underline = underline

        if self.view():
            w = round(self._Size[0] * self.view().width())
            h = round(self._Size[1] * self.view().height())

            pa = QPainter()

            # 字体
            font_size = 16
            font = QFont(self._TextFont, font_size)
            font.setPointSizeF(font_size)
            font.setBold(self._Bold)
            font.setItalic(self._Italic)
            font.setUnderline(self._Underline)

            # 计算字体大小
            image = QImage(w, h, QImage.Format.Format_RGB32)

            pa.begin(image)
            pa.setPen(self._ForeColor)
            # for s in range(2,18):
            pa.setFont(font)
            rect = pa.drawText(0, 0, w, h, getattr(Qt, self._TextAlign), self._Text)
            if rect.width() > 0 and rect.height() > 0:
                if 0.5 * h / rect.height() < (w - 20) / rect.width():
                    font_size *= 0.5 * h / rect.height()
                else:
                    font_size *= (w - 20) / rect.width()
                font.setPointSizeF(font_size)
            pa.end()

            # 绘制文本
            image = QImage(w, h, QImage.Format.Format_RGB32)
            pa.begin(image)
            pa.setFont(font)
            pa.setPen(self._ForeColor)
            pa.fillRect(0, 0, w, h, self._BackColor)
            pa.drawText(0, 0, w, h, getattr(Qt, self._TextAlign), self._Text)
            pa.end()

            self.draw(image=image, size=size, pos=pos, anchor=anchor)

    def __setstate__(self, state):
        self.__init__()

        for kw in state['vtk']:
            vmi.vtkInstance_sets(getattr(self, kw), state['vtk'][kw])

        for kw in state['set']:
            getattr(self, kw.replace('_', 'set', 1))(state['set'][kw])

        self.draw_text(text=state['draw']['_Text'], text_font=state['draw']['_TextFont'],
                       text_align=state['draw']['_TextAlign'], fore_color=state['draw']['_ForeColor'],
                       back_color=state['draw']['_BackColor'], bold=state['draw']['_Bold'],
                       italic=state['draw']['_Italic'], underline=state['draw']['_Underline'],
                       size=state['draw']['_Size'], pos=state['draw']['_Pos'], anchor=state['draw']['_Anchor'])

        del state['vtk']
        del state['draw']
        del state['set']
        self.__dict__.update(state)
        self.setView(self._View)

    def __getstate__(self):
        state = self.__dict__.copy()
        state = {kw: state[kw] for kw in state if kw not in ['menu', 'actions', '__METAOBJECT__']}

        state_set = {kw: getattr(self, kw) for kw in ['_Pickable', '_Visible']}
        state = {kw: state[kw] for kw in state if kw not in state_set}

        state_draw = {kw: getattr(self, kw) for kw in
                      ['_Anchor', '_BackColor', '_Bold', '_ForeColor', '_Italic', '_Pos', '_Size', '_Text',
                       '_TextAlign', '_TextFont', '_Underline']}

        state_vtk = {}
        for kw in state:
            if isinstance(state[kw], vtk.vtkObject):
                if not hasattr(vmi, 'pickle_' + state[kw].__class__.__name__):
                    state_vtk[kw] = vmi.vtkInstance_gets(state[kw])
        state = {kw: state[kw] for kw in state if kw not in state_vtk}

        state['set'] = state_set
        state['draw'] = state_draw
        state['vtk'] = state_vtk
        return state

    def __repr__(self):
        state = self.__getstate__()
        state = {**state['set'], **state['draw']}
        return pprint.pformat(state, width=400, compact=True)

    def setView(self, view) -> None:
        "设置视图，同时更新数据关联"
        if self._View:
            self._View.renderer().RemoveActor(self._Prop)
        self._View = view
        if self._View:
            self._View.renderer().AddActor(self._Prop)
        self.bindData()
        self.draw_text()

    def resize(self) -> None:
        self.draw_text()


class ImageSlice(QObject, vmi.Menu, vmi.Mouse):
    """图像断层，将vtkImageData通过vtkImageResliceMapper/vtkImageSlice显示"""

    def __init__(self, view: vmi.View = None, visible: bool = True, pickable: bool = True,
                 colorWindow: List[int] = (350, 50)):
        """
        :param view: 视图
        :param visible: 可见
        :param pickable: 可交互
        :param colorWindow: 窗宽窗位 [width, level]
        """
        QObject.__init__(self)
        vmi.Menu.__init__(self)

        self.actions = {'SlicePlaneValueNormal': QAction(''),
                        'SlicePlaneAxial': QAction('横断位 (Axial)'),
                        'SlicePlaneSagittal': QAction('矢状位 (Sagittal)'),
                        'SlicePlaneCoronal': QAction('冠状位 (Coronal)'),
                        'WindowValue': QAction(''),
                        'WindowAuto': QAction('自动 (Auto)'),
                        'WindowBone': QAction('骨骼 (Bone)'),
                        'WindowSoft': QAction('组织 (Soft)')}

        self.actions['SlicePlaneAxial'].triggered.connect(self.setSlicePlaneNormal_Axial)
        self.actions['SlicePlaneSagittal'].triggered.connect(self.setSlicePlaneNormal_Sagittal)
        self.actions['SlicePlaneCoronal'].triggered.connect(self.setSlicePlaneNormal_Coronal)
        self.actions['WindowAuto'].triggered.connect(self.setColorWindow_Auto)
        self.actions['WindowBone'].triggered.connect(self.setColorWindow_Bone)
        self.actions['WindowSoft'].triggered.connect(self.setColorWindow_Soft)

        def aboutToShow():
            self.actions['SlicePlaneValueNormal'].setText(
                '法向 (Normal)' + ' = ' + repr(tuple(self._SlicePlaneNormal)))
            self.actions['WindowValue'].setText(
                '宽/位 (W/L)' + ' = ' + (repr(tuple(self._ColorWindow))))

            self.menu.clear()

            menu = QMenu('切面 (Slice plane)')
            menu.addAction(self.actions['SlicePlaneValueNormal'])
            menu.addSeparator()
            menu.addAction(self.actions['SlicePlaneAxial'])
            menu.addAction(self.actions['SlicePlaneSagittal'])
            menu.addAction(self.actions['SlicePlaneCoronal'])
            self.menu.addMenu(menu)

            menu = QMenu('窗 (Window)')
            menu.addAction(self.actions['WindowValue'])
            menu.addSeparator()
            menu.addAction(self.actions['WindowAuto'])
            menu.addAction(self.actions['WindowBone'])
            menu.addAction(self.actions['WindowSoft'])
            self.menu.addMenu(menu)

        self.menu.aboutToShow.connect(aboutToShow)

        vmi.Mouse.__init__(self, menu=self)
        self.mouse['NoButton']['Wheel'] = [self.slice]
        self.mouse['LeftButton']['PressMove'] = [self.windowMove]

        self._Mapper = vtk.vtkImageResliceMapper()
        self._Prop = vtk.vtkImageSlice()
        self._Prop._Prop = self

        self._Property = self._Prop.GetProperty()
        self._Property.SetInterpolationTypeToCubic()
        self._Property.SetUseLookupTableScalarRange(1)

        self._SlicePlane = vtk.vtkPlane()
        self._Mapper.SetSlicePlane(self._SlicePlane)

        self._Data = vtk.vtkImageData()
        self._BindData = self

        self._SlicePlaneOrigin = np.array([0, 0, 0])
        self._BindSlicePlaneOrigin = [self]

        self._SlicePlaneNormal = np.array([0, 0, 1])
        self._BindSlicePlaneNormal = [self]

        self._LookupTable = vtk.vtkLookupTable()
        self._BindColorWindow = [self]

        self._View = self._Visible = self._Pickable = None
        self._ColorWindow = [350, 50]
        self.setView(view)
        self.setVisible(visible)
        self.setPickable(pickable)
        self.setColorWindow(colorWindow)
        self.setSlicePlane(self._SlicePlaneOrigin, self._SlicePlaneNormal)

    def __setstate__(self, state):
        self.__init__()

        for kw in state['vtk']:
            vmi.vtkInstance_sets(getattr(self, kw), state['vtk'][kw])

        for kw in state['set']:
            getattr(self, kw.replace('_', 'set', 1))(state['set'][kw])

        del state['vtk']
        del state['set']
        self.__dict__.update(state)
        self.setView(self._View)

    def __getstate__(self):
        state = self.__dict__.copy()
        state = {kw: state[kw] for kw in state if kw not in ['menu', 'actions', '__METAOBJECT__']}

        state_set = {kw: getattr(self, kw) for kw in [
            '_Data', '_Pickable', '_Visible', '_ColorWindow', '_LookupTable', '_SlicePlaneOrigin', '_SlicePlaneNormal']}
        state = {kw: state[kw] for kw in state if kw not in state_set}

        state_vtk = {}
        for kw in state:
            if isinstance(state[kw], vtk.vtkObject):
                if not hasattr(vmi, 'pickle_' + state[kw].__class__.__name__):
                    state_vtk[kw] = vmi.vtkInstance_gets(state[kw])
        state = {kw: state[kw] for kw in state if kw not in state_vtk}

        state['set'] = state_set
        state['vtk'] = state_vtk
        return state

    def __repr__(self):
        state = self.__getstate__()['set']
        state['_Data'] = {'GetActualMemorySize': self.data().GetActualMemorySize(),
                          'GetBounds': self.data().GetBounds(),
                          'GetCenter': self.data().GetCenter(),
                          'GetExtent': self.data().GetExtent(),
                          'GetNumberOfCells': self.data().GetNumberOfCells(),
                          'GetNumberOfPoints': self.data().GetNumberOfPoints(),
                          'GetOrigin': self.data().GetOrigin(),
                          'GetScalarRange': self.data().GetScalarRange(),
                          'GetSpacing': self.data().GetSpacing()}
        return pprint.pformat(state, width=400, compact=True)

    def view(self) -> vmi.View:
        """返回视图"""
        return self._View

    def setView(self, view) -> None:
        """设置视图"""
        if self._View:
            self._View.renderer().RemoveActor(self._Prop)
        self._View = view
        if self._View:
            self._View.renderer().AddActor(self._Prop)
        self.bindData()

    def updateInTime(self) -> None:
        """及时更新"""
        if self._View:
            self._View.updateInTime()

    def visible(self) -> bool:
        """返回可见状态"""
        return self._Visible

    def setVisible(self, visible: bool) -> None:
        """设置可见状态"""
        self.updateInTime()
        self._Visible = visible
        self._Prop.SetVisibility(1 if visible else 0)

    def visibleToggle(self) -> None:
        """切换可见状态"""
        return self.setVisible(not self.visible())

    def pickable(self) -> bool:
        """返回可交互状态"""
        return self._Pickable

    def setPickable(self, pickable: bool) -> None:
        """设置可交互状态"""
        self.updateInTime()
        self._Pickable = pickable
        self._Prop.SetPickable(1 if self._Pickable else 0)

    def data(self) -> vtk.vtkImageData:
        """返回数据"""
        return self._Data

    def slicePlane(self) -> vtk.vtkPlane:
        """返回断层平面"""
        return self._SlicePlane

    def slicePlaneOrigin(self) -> np.ndarray:
        return self._SlicePlaneOrigin

    def slicePlaneNormal(self) -> np.ndarray:
        return self._SlicePlaneNormal

    def lookupTable(self) -> vtk.vtkLookupTable:
        """返回颜色查找表"""
        return self._LookupTable

    def bindData(self, other=None) -> None:
        """关联数据到视图，可以关联其它对象，other = self 关联自身对象，other = None 关联当前对象"""
        self.updateInTime()
        if hasattr(other, '_Data'):
            self._BindData = other
        self._Mapper.SetInputData(self._BindData.data())
        self._Prop.SetMapper(self._Mapper)

    def bindSlicePlaneOrigin(self, others: List) -> None:
        """关联断层平面的原点，设置列表内的所有对象彼此关联，others = [] 取消关联"""
        binds = others.copy()
        if self not in binds:
            binds.append(self)
        for i in binds:
            i._BindSlicePlaneOrigin = binds.copy()
            i.setSlicePlaneOrigin(self._SlicePlaneOrigin)

    def bindSlicePlaneNormal(self, others: List) -> None:
        """关联断层平面的原点，设置列表内的所有对象彼此关联，others = [] 取消关联"""
        binds = others.copy()
        if self not in binds:
            binds.append(self)
        for i in binds:
            i._BindSlicePlaneNormal = binds.copy()
            i.setSlicePlaneNormal(self._SlicePlaneNormal)

    def bindColorWindow(self, others: List) -> None:
        """关联窗宽窗位，设置列表内的所有对象彼此关联，others = [] 取消关联"""
        binds = others.copy()
        if self not in binds:
            binds.append(self)
        for i in binds:
            i._BindColorWindow = binds.copy()
            i.setColorWindow(self._ColorWindow)

    def setLookupTable(self, lookup_table) -> None:
        """设置颜色查找表"""
        self._LookupTable = lookup_table
        self._Property.SetLookupTable(self.lookupTable())
        self.updateInTime()

    def setData(self, data) -> None:
        """设置数据，可以直接设置其它同类实例，或vtkImageData"""
        self.updateInTime()
        if hasattr(data, '_Data'):
            self.data().ShallowCopy(data.data())
        elif isinstance(data, self.data().__class__):
            self.data().ShallowCopy(data)
        else:
            self.data().ShallowCopy(self.data().__class__())
        self.bindData()

    def delete(self) -> None:
        """清除视图关联，在del之前调用"""
        self.updateInTime()
        self.data().ReleaseData()
        self.setView(None)

    def setSlicePlane(self, origin: Union[np.ndarray, List[float]], normal: Union[np.ndarray, List[float]]) -> None:
        """
        设置断层平面的原点和法向

        :param origin: 原点
        :param normal: 法向
        :return: None
        """
        for i in self._BindSlicePlaneOrigin:
            i._SlicePlaneOrigin = np.array(origin)
            i._SlicePlaneNormal = np.array(normal)
            self.slicePlane().SetOrigin(list(origin))
            self.slicePlane().SetNormal(list(normal))
            self.updateInTime()

    def setSlicePlaneOrigin(self, origin: Union[np.ndarray, List[float]]) -> None:
        """
        设置断层平面的原点，法向不变

        :param origin: 原点
        :return: None
        """
        for i in self._BindSlicePlaneOrigin:
            i._SlicePlaneOrigin = np.array(origin)
            i.slicePlane().SetOrigin(list(origin))
            i.updateInTime()

    def setSlicePlaneNormal(self, normal: Union[np.ndarray, List[float]]) -> None:
        """
        设置断层平面的法向，原点不变

        :param normal: 法向
        :return: None
        """
        for i in self._BindSlicePlaneNormal:
            i._SlicePlaneNormal = np.array(normal)
            i.slicePlane().SetNormal(normal)
            i.updateInTime()

    def setSlicePlaneOrigin_Center(self) -> None:
        """设置断层平面原点为数据中心"""
        self.setSlicePlaneOrigin(self._BindData.data().GetCenter())

    def setSlicePlaneNormal_Axial(self) -> None:
        """设置断层平面法向为横断位"""
        self.setSlicePlaneNormal([0, 0, 1])

    def setSlicePlaneNormal_Sagittal(self) -> None:
        """设置断层平面法向为矢状位"""
        self.setSlicePlaneNormal([-1, 0, 0])

    def setSlicePlaneNormal_Coronal(self) -> None:
        """设置断层平面法向为冠状位"""
        self.setSlicePlaneNormal([0, -1, 0])

    def setInterpolationType_Nearest(self) -> None:
        """设置图像显示的插值方式为邻位插值"""
        self._Property.SetInterpolationTypeToNearest()
        self.updateInTime()

    def setInterpolationType_Linear(self) -> None:
        """设置图像显示的插值方式为线性插值"""
        self._Property.SetInterpolationTypeToLinear()
        self.updateInTime()

    def setInterpolationType_Cubic(self) -> None:
        """设置图像显示的插值方式为三次插值"""
        self._Property.SetInterpolationTypeToCubic()
        self.updateInTime()

    def setColorWindow(self, color_window: List[int]) -> None:
        """设置窗宽窗位"""
        width = min(max(round(color_window[0]), 0), 20000)
        level = min(max(round(color_window[1]), -10000), 10000)

        self._ColorWindow[0] = int(width)
        self._ColorWindow[1] = int(level)

        r = [self._ColorWindow[1] - 0.5 * self._ColorWindow[0],
             self._ColorWindow[1] + 0.5 * self._ColorWindow[0]]

        t = vtk.vtkLookupTable()
        t.SetNumberOfTableValues(self._ColorWindow[0])
        t.SetTableRange(r)
        t.SetBelowRangeColor(0, 0, 0, 1.0)
        t.SetAboveRangeColor(1, 1, 1, 1.0)
        t.SetUseBelowRangeColor(1)
        t.SetUseAboveRangeColor(1)

        for i in range(self._ColorWindow[0]):
            v = i / self._ColorWindow[0]
            t.SetTableValue(i, (v, v, v, 1.0))

        for i in self._BindColorWindow:
            i._ColorWindow[0] = self._ColorWindow[0]
            i._ColorWindow[1] = self._ColorWindow[1]
            i.setLookupTable(t)

    def setColorWindowWidth(self, width: int) -> None:
        """设置窗宽，窗位不变"""
        width = min(max(width, 0), 20000)
        self.setColorWindow([width, self._ColorWindow[1]])

    def setColorWindowLevel(self, level: int) -> None:
        """设置窗位，窗宽不变"""
        level = min(max(level, -10000), 10000)
        self.setColorWindow([self._ColorWindow[0], level])

    def setColorWindow_Auto(self) -> None:
        """设置窗宽窗位为适应全局"""
        r = vtk.vtkImageHistogramStatistics()
        r.SetInputData(self._BindData.data())
        r.SetAutoRangePercentiles(1, 99)
        r.SetGenerateHistogramImage(0)
        r.Update()
        r = r.GetAutoRange()
        self.setColorWindow([round(r[1] - r[0]), round(0.5 * (r[0] + r[1]))])

    def setColorWindow_Bone(self) -> None:
        """设置窗宽窗位为适应骨骼[1000, 400]"""
        self.setColorWindow([1000, 400])

    def setColorWindow_Soft(self) -> None:
        """设置窗宽窗位为适应软组织[350, 50]"""
        self.setColorWindow([350, 50])

    def slice(self, **kwargs):
        o = list(self.slicePlane().GetOrigin())
        n = self.slicePlane().GetNormal()
        dim = self._BindData.data().GetDimensions()
        dxyz = self._BindData.data().GetSpacing()

        n = [abs(_) / (n[0] ** 2 + n[1] ** 2 + n[2] ** 2) ** 0.5 for _ in n]
        dn = n[0] * dxyz[0] + n[1] * dxyz[1] + n[2] * dxyz[2]

        for i in range(3):
            if dim[i] > 1:
                o[i] += kwargs['delta'] * dn * n[i]

        if vtk.vtkMath.PlaneIntersectsAABB(self._BindData.data().GetBounds(), n, o) == 0:
            self.setSlicePlaneOrigin(o)

    def windowMove(self, **kwargs):
        if self.view():
            dx, dy = self.view().pickVt_Display()
            width = self._ColorWindow[0] + 2 * dx
            level = self._ColorWindow[1] - 2 * dy
            self.setColorWindow([width, level])


class ImageVolume(QObject, vmi.Menu, vmi.Mouse):
    """图像立体，将vtkImageData通过vtkGPUVolumeRayCastMapper/vtkVolume显示"""

    def __init__(self, view: vmi.View = None, visible: bool = True, pickable: bool = False,
                 color: Dict[int, List[float]] = None, opacityScalar: Dict[int, float] = None,
                 opacityGradient: Dict[int, float] = None):
        """
        :param view: 视图
        :param visible: 可见
        :param pickable: 可交互
        :param color: 颜色方案 {scalar: [red, green, blue]}
        :param opacityScalar: 不透明度方案 {scalar: opacity}
        :param opacityGradient: 不透明度梯度方案 {scalar: opacity}
        """
        QObject.__init__(self)
        vmi.Menu.__init__(self)

        self.actions = {'Threshold': QAction('阈值 (Threshold)')}
        self.actions['Threshold'].triggered.connect(self.setOpacity_Threshold)

        def aboutToShow():
            self.menu.clear()

            menu = QMenu('风格 (Style)')
            menu.addAction(self.actions['Threshold'])
            self.menu.addMenu(menu)

        self.menu.aboutToShow.connect(aboutToShow)

        vmi.Mouse.__init__(self)

        self._Mapper = vtk.vtkGPUVolumeRayCastMapper()
        self._Prop = vtk.vtkVolume()
        self._Prop._Prop = self

        self._Mapper.SetBlendModeToComposite()
        self._Mapper.SetMaxMemoryInBytes(4096)
        self._Mapper.SetMaxMemoryFraction(1)
        self._Mapper.SetAutoAdjustSampleDistances(0)
        self._Mapper.SetLockSampleDistanceToInputSpacing(1)
        self._Mapper.SetUseJittering(1)

        self._Property = self._Prop.GetProperty()
        self._Property.SetInterpolationTypeToLinear()
        self._Property.SetAmbient(0)
        self._Property.SetDiffuse(1)
        self._Property.SetShade(1)

        self._Data = vtk.vtkImageData()
        self._BindData = self

        self._View = self._Visible = self._Pickable = None
        self._Color = self._OpacityScalar = self._OpacityGradient = None

        self.setView(view)
        self.setVisible(visible)
        self.setPickable(pickable)
        self.setColor(color if color else {0: [1, 1, 1]})
        self.setOpacityScalar(opacityScalar if opacityScalar else {0: 0, 400: 1})
        self.setOpacityGradient(opacityGradient if opacityGradient else {})

    def __setstate__(self, state):
        self.__init__()

        for kw in state['vtk']:
            vmi.vtkInstance_sets(getattr(self, kw), state['vtk'][kw])

        for kw in state['set']:
            getattr(self, kw.replace('_', 'set', 1))(state['set'][kw])

        del state['vtk']
        del state['set']
        self.__dict__.update(state)
        self.setView(self._View)

    def __getstate__(self):
        state = self.__dict__.copy()
        state = {kw: state[kw] for kw in state if kw not in ['menu', 'actions', '__METAOBJECT__']}

        state_set = {kw: getattr(self, kw) for kw in
                     ['_Data', '_Pickable', '_Visible', '_Color', '_OpacityScalar', '_OpacityGradient']}
        state = {kw: state[kw] for kw in state if kw not in state_set}

        state_vtk = {}
        for kw in state:
            if isinstance(state[kw], vtk.vtkObject):
                if not hasattr(vmi, 'pickle_' + state[kw].__class__.__name__):
                    state_vtk[kw] = vmi.vtkInstance_gets(state[kw])
        state = {kw: state[kw] for kw in state if kw not in state_vtk}

        state['set'] = state_set
        state['vtk'] = state_vtk
        return state

    def __repr__(self):
        state = self.__getstate__()['set']
        state['_Data'] = {'GetActualMemorySize': self.data().GetActualMemorySize(),
                          'GetBounds': self.data().GetBounds(),
                          'GetCenter': self.data().GetCenter(),
                          'GetExtent': self.data().GetExtent(),
                          'GetNumberOfCells': self.data().GetNumberOfCells(),
                          'GetNumberOfPoints': self.data().GetNumberOfPoints(),
                          'GetOrigin': self.data().GetOrigin(),
                          'GetScalarRange': self.data().GetScalarRange(),
                          'GetSpacing': self.data().GetSpacing()}
        return pprint.pformat(state, width=400, compact=True)

    def view(self) -> vmi.View:
        """返回视图"""
        return self._View

    def setView(self, view) -> None:
        """设置视图"""
        if self._View:
            self._View.renderer().RemoveActor(self._Prop)
        self._View = view
        if self._View:
            self._View.renderer().AddActor(self._Prop)
        self.bindData()

    def updateInTime(self) -> None:
        """及时刷新视图"""
        if self._View:
            self._View.updateInTime()

    def visible(self) -> bool:
        """返回可见状态"""
        return self._Visible

    def setVisible(self, visible: bool) -> None:
        """设置可见状态"""
        self.updateInTime()
        self._Visible = visible
        self._Prop.SetVisibility(1 if visible else 0)

    def visibleToggle(self) -> None:
        """切换可见状态"""
        return self.setVisible(not self.visible())

    def pickable(self) -> bool:
        """返回可交互状态"""
        return self._Pickable

    def setPickable(self, pickable: bool) -> None:
        """设置可交互状态"""
        self.updateInTime()
        self._Pickable = pickable
        self._Prop.SetPickable(1 if self._Pickable else 0)

    def data(self) -> vtk.vtkImageData:
        """返回数据"""
        return self._Data

    def delete(self) -> None:
        """清除视图关联，在del之前调用"""
        self.updateInTime()
        image = vtk.vtkImageData()
        image.SetExtent(0, 0, 0, 0, 0, 0)
        image.AllocateScalars(vtk.VTK_SHORT, 1)
        self.setData(image)
        self.setView(None)

    def bindData(self, other=None) -> None:
        """关联数据到视图，可以关联其它对象，other = self 关联自身对象，other = None 重新关联当前对象"""
        if hasattr(other, '_Data'):
            self._BindData = other
        self._Mapper.SetInputData(self._BindData.data())
        self._Prop.SetMapper(self._Mapper)
        self.updateInTime()

    def setData(self, data) -> None:
        """设置数据，可以直接设置其它同类实例，或vtkImageData"""
        self.updateInTime()
        if hasattr(data, 'data'):
            self.data().DeepCopy(data.data())
        elif isinstance(data, self.data().__class__):
            self.data().DeepCopy(data)
        else:
            self.data().DeepCopy(self.data().__class__())
        self.bindData()

    def setColor(self, color: Dict[int, List[float]]) -> None:
        """设置颜色方案 {scalar: [red, green, blue]}"""
        self.updateInTime()
        self._Color = color
        f = vtk.vtkColorTransferFunction()
        for x in self._Color:
            r, g, b = self._Color[x]
            f.AddRGBPoint(x, r, g, b)
        self._Property.SetColor(f)

    def setOpacityScalar(self, opacity_scalar: Dict[int, float]) -> None:
        self.updateInTime()
        self._OpacityScalar = opacity_scalar
        f = vtk.vtkPiecewiseFunction()
        for x in self._OpacityScalar:
            f.AddPoint(x, self._OpacityScalar[x])
        self._Property.SetScalarOpacity(f)

    def setOpacity_Threshold(self, threshold: int = None):
        """设置不透明度方案为阈值方案，threshold以上显示，threshold以下完全透明"""
        self.setOpacityScalar({threshold - 1: 0, threshold: 1})

    def setOpacityGradient(self, opacity_gradient: Dict[int, float]) -> None:
        """设置不透明度梯度方案，没啥用"""
        self.updateInTime()
        self._OpacityGradient = opacity_gradient
        f = vtk.vtkPiecewiseFunction()
        for x in self._OpacityGradient:
            f.AddPoint(x, self._OpacityGradient[x])
        self._Property.SetGradientOpacity(f)



import copyreg
import pathlib
import tempfile
from typing import Dict, Any

import numpy as np
import vtk
from OCC.Core.IFSelect import *
from OCC.Core.STEPControl import *
from OCC.Core.TopoDS import *
from PySide2.QtCore import *
from PySide2.QtGui import *

import vmi

vtkget_ignore = ('GetDebug', 'GetGlobalReleaseDataFlag', 'GetGlobalWarningDisplay', 'GetReferenceCount',
                 'GetAAFrames', 'GetFDFrames', 'GetSubFrames', 'GetUseConstantFDOffsets', 'GetStereoCapableWindow',
                 'GetForceCompileOnly', 'GetGlobalImmediateModeRendering', 'GetImmediateModeRendering',
                 'GetScalarMaterialMode', 'GetReleaseDataFlag')


def vtkInstance_gets(instance) -> Dict[str, Any]:
    """
    返回vtk类型实例的所有Set/Get状态值

    :param instance: vtk类型实例
    :return: 状态值{'GetXXX': Any}
    """
    gets = {}
    for get_name in dir(instance):
        if get_name.startswith('Get'):
            set_name = get_name.replace('Get', 'Set', 1)
            if hasattr(instance, set_name) and get_name not in vtkget_ignore:
                try:
                    a = getattr(instance, get_name)()
                    if 'vtk' not in str(type(a)):
                        gets[get_name] = a
                except TypeError:
                    pass
    return gets


def vtkInstance_sets(instance, gets) -> None:
    """
    设置vtk类型实例的所有Set/Get状态值
    :param instance: vtk类型实例
    :param gets: 状态值{'GetXXX': Any}
    :return: None
    """
    for get_name in gets:
        set_name = get_name.replace('Get', 'Set', 1)
        try:
            getattr(instance, set_name)(gets[get_name])
        except TypeError:
            pass


def pickle_vtkImageData(instance: vtk.vtkImageData):
    gets = vtkInstance_gets(instance)
    if instance.GetNumberOfPoints() == 0:
        data = bytes()
    else:
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.vti'
            vmi.imSaveFile_XML(instance, str(p))
            data = p.read_bytes()
    return unpickle_vtkImageData, (gets, data,)


def unpickle_vtkImageData(gets, data):
    if len(data) == 0:
        instance = vtk.vtkImageData()
    else:
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.vti'
            p.write_bytes(data)
            instance = vmi.imOpenFile_XML(str(p))
    vtkInstance_sets(instance, gets)
    return instance


def pickle_vtkLookupTable(instance: vtk.vtkLookupTable):
    gets = vtkInstance_gets(instance)
    data = []
    for i in range(instance.GetNumberOfTableValues()):
        data.append(instance.GetTableValue(i))
    return unpickle_vtkLookupTable, (gets, data,)


def unpickle_vtkLookupTable(gets, data):
    instance = vtk.vtkLookupTable()
    vtkInstance_sets(instance, gets)
    instance.SetNumberOfTableValues(len(data))
    for i in range(len(data)):
        instance.SetTableValue(i, data[i])
    instance.Build()
    return instance


def pickle_vtkMatrix4x4(instance: vtk.vtkMatrix4x4):
    gets = vtkInstance_gets(instance)
    data = np.zeros((4, 4))
    for j in range(4):
        for i in range(4):
            data[i, j] = instance.GetElement(j, i)
    return unpickle_vtkMatrix4x4, (gets, data,)


def unpickle_vtkMatrix4x4(gets, data):
    instance = vtk.vtkMatrix4x4()
    vtkInstance_sets(instance, gets)
    for j in range(4):
        for i in range(4):
            instance.SetElement(j, i, data[i, j])
    return instance


def pickle_vtkPolyData(instance: vtk.vtkPolyData):
    gets = vtkInstance_gets(instance)
    if instance.GetNumberOfPoints() == 0:
        data = bytes()
    else:
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.vtp'
            vmi.pdSaveFile_XML(instance, str(p))
            data = p.read_bytes()
    return unpickle_vtkPolyData, (gets, data,)


def unpickle_vtkPolyData(gets, data):
    if len(data) == 0:
        instance = vtk.vtkPolyData()
    else:
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.vtp'
            p.write_bytes(data)
            instance = vmi.pdOpenFile_XML(str(p))
    vtkInstance_sets(instance, gets)
    return instance


copyreg.pickle(vtk.vtkImageData, pickle_vtkImageData)
copyreg.pickle(vtk.vtkLookupTable, pickle_vtkLookupTable)
copyreg.pickle(vtk.vtkMatrix4x4, pickle_vtkMatrix4x4)
copyreg.pickle(vtk.vtkPolyData, pickle_vtkPolyData)


def pickle_TopoDS_Shape(instance: TopoDS_Shape):
    if instance.IsNull():
        data = bytes()
    else:
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.stp'
            w = STEPControl_Writer()
            w.Transfer(instance, STEPControl_AsIs)
            status = w.Write(str(p))
            if status == IFSelect_RetDone:
                data = p.read_bytes()
            else:
                data = bytes()
    return unpickle_TopoDS_Shape, (data,)


def unpickle_TopoDS_Shape(data):
    if len(data) == 0:
        instance = TopoDS_Shape()
    else:
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.stp'
            p.write_bytes(data)
            r = STEPControl_Reader()
            status = r.ReadFile(str(p))
            if status == IFSelect_RetDone:
                r.TransferRoots()
                instance = r.OneShape()
            else:
                instance = TopoDS_Shape()
    return instance


copyreg.pickle(TopoDS_Shape, pickle_TopoDS_Shape)


def pickle_QImage(instance: QImage):
    ba = QByteArray()
    buffer = QBuffer(ba)
    buffer.open(QIODevice.WriteOnly)
    instance.save(buffer, 'PNG')
    return unpickle_QImage, (ba.data(), instance.format().name.decode())


def unpickle_QImage(data, format):
    ba = QByteArray(data)
    instance = QImage.fromData(ba, format)
    return instance


copyreg.pickle(QImage, pickle_QImage)


def contains_zh_CN(text) -> bool:
    """判断字符串是否包含中文"""
    return text.encode() != text.encode('cp936')


def read_Medraw(file_name: str) -> Dict:
    """
    读取.medraw文件，输出

    :param file_name: 文件路径
    :return: {'appName', 'appVersion', 'origin', 'spacing', 'extent', 'aryRef', 'aryEdi', 'imageRef', 'imageEdi'}
    """
    f = QFile(file_name)
    f.open(QIODevice.ReadOnly)
    s = QDataStream(f)

    appName = s.readQString()
    appVersion = s.readQString()
    origin = [s.readDouble(), s.readDouble(), s.readDouble()]
    spacing = [s.readDouble(), s.readDouble(), s.readDouble()]
    extent = [s.readInt32(), s.readInt32(), s.readInt32(), s.readInt32(), s.readInt32(), s.readInt32()]
    shape = [extent[5] - extent[4] + 1, extent[3] - extent[2] + 1, extent[1] - extent[0] + 1]

    aryRef = np.zeros(shape)
    aryEdi = np.zeros(shape)

    for k in range(aryRef.shape[0]):
        for j in range(aryRef.shape[1]):
            for i in range(aryRef.shape[2]):
                aryRef[k][j][i] = s.readInt16()
                aryEdi[k][j][i] = s.readUInt8()

    imageRef = vmi.imVTK_Array(aryRef, origin, spacing)
    imageEdi = vmi.imVTK_Array(aryEdi, origin, spacing)

    return {'appName': appName, 'appVersion': appVersion, 'origin': origin, 'spacing': spacing, 'extent': extent,
            'aryRef': aryRef, 'aryEdi': aryEdi, 'imageRef': imageRef, 'imageEdi': imageEdi}
